#include "OutputISGProperties.h"
#include "Configuration.h"
#include "GenericTopology.h" 
#include "ScalarStructure.h"
#include "topologyutilities.h"
#include "OutputCache.h" 
#include "inputValueDefinitions.h"

#include <iomanip>

using namespace ProtoMol::Report;

using std::string;
using std::vector;
using std::setw;
using std::endl;
using std::flush;
using std::setprecision;
using std::setiosflags;
using std::resetiosflags;
using std::ofstream;

namespace ProtoMol {
  //________________________________________________________ Output
  const string  OutputISGProperties::keyword("iSGPropertiesfile");

  OutputISGProperties::OutputISGProperties():OutputFile(),myDoMolecularTemperature(true){}
  OutputISGProperties::OutputISGProperties(const string& filename, int freq, bool doMolTemp):OutputFile(filename,freq),myDoMolecularTemperature(doMolTemp){}

  void OutputISGProperties::doInitialize(){
    ofstream iSGPropertiesHeaderFile(string(myFilename + ".header").c_str(), std::ios::out | std::ios::trunc);
    if(!iSGPropertiesHeaderFile)
      report << error <<" Can not open \'"<<myFilename<<".header\' for "<<getId()<<"."<<endr;

    iSGPropertiesHeaderFile << setw(18)
			    << "Time(fs)" << " "
			    << setw(14)
                            << "E_kinetic" << " "
                            << setw(14)
			    << "E_potential" << " "
			    << setw(14)
			    << "Temperature" << " ";
    if(myEnergies->virial())
      iSGPropertiesHeaderFile << setw(14)
			      << "Pressure(bar)" << " ";
			      
    if(myEnergies->molecularVirial())
      iSGPropertiesHeaderFile << setw(14)
			      << "Mol_Pres(bar)" << " ";
    iSGPropertiesHeaderFile << setw(12)
			    << "Volume(A^3)";
    for (unsigned int i=0; i<myTopology->iSGNumMols.size(); i++) {
      iSGPropertiesHeaderFile << setw(9)
			      << "N[" << i << "]";}
    iSGPropertiesHeaderFile << setw(12) 
			    << "deltaCQ"
			    << setw(14)
			    << "Steps" << endl;
    
    iSGPropertiesHeaderFile.close();
    open();
    close();

  }

  void OutputISGProperties::doRun(int){

    if (myEnergies->output()) {

      if(!reopen())
	report << error <<" Can not open \'"<<myFilename<<"\' for "<<getId()<<"."<<endr;

      myFile  << resetiosflags(std::ios::showpoint |  std::ios::fixed | std::ios::floatfield)
	      << setw(18)
	      << setprecision(2)
	      << setiosflags(std::ios::showpoint | std::ios::fixed)
	      << myCache->time() << " "
	      << resetiosflags(std::ios::showpoint |  std::ios::fixed | std::ios::floatfield)
	      << setiosflags(std::ios::floatfield)
	      << setprecision(8)
	      << setw(14)
              << myCache->kineticEnergy()
              << setw(14)
              << myCache->potentialEnergy()
	      << setw(14)
	      << myCache->temperature()
	      << setw(14)
	      << myCache->pressure()
	      << setw(14)
	      << myCache->molecularPressure()
	      << setw(15)
	      << myCache->volume() << "   ";
      for (unsigned int i=0; i<myTopology->iSGNumMols.size(); i++) {
	myFile << setw(10)
	       << myTopology->iSGNumMols[i];}
      myFile  << resetiosflags(std::ios::showpoint |  std::ios::fixed | std::ios::floatfield)
	      << setiosflags(std::ios::floatfield)
	      << setprecision(8)  
	      << setw(16) 
	      << (*myEnergies)[ScalarStructure::CQFLUCTUATION]
	      << setw(10)
	      << (*myEnergies)[ScalarStructure::DELTATIME] << endl;
      
      reclose();

    }

  }

  void OutputISGProperties::doFinalize(int){
    close();
  }

  Output* OutputISGProperties::doMake(string&, const vector<Value>& values) const{
    return (new OutputISGProperties(values[0],values[1],values[2]));
  }

  void OutputISGProperties::getParameters(vector<Parameter> &parameter) const{
    parameter.push_back(Parameter(getId(),Value(myFilename,ConstraintValueType::NotEmpty())));
    parameter.push_back(Parameter("energiesOutputFreq",Value(myOutputFreq,ConstraintValueType::Positive())));
    parameter.push_back(Parameter("molecularTemperature",Value(myDoMolecularTemperature),true));
  }

  bool OutputISGProperties::adjustWithDefaultParameters(vector<Value>& values, const Configuration* config) const{
    if(!checkParameterTypes(values))
      return false;
    if(config->valid(InputOutputfreq::keyword) && !values[1].valid())
      values[1] = (*config)[InputOutputfreq::keyword];
    return checkParameters(values);
  }

}
