#include "ModifierISG.h"
#include "mathutilities.h"
#include "ScalarStructure.h" 
#include "GenericTopology.h" 
#include "Report.h"
#include "Integrator.h"
#include "iSGIntegrator.h"
#include "XSCReader.h"
#include <string>
#include <vector>

using std::vector;
using std::map;
using std::string;

using namespace ProtoMol::Report;

namespace ProtoMol {

  //_________________________________________________________________ ModifierISG
  ModifierISG::ModifierISG(iSGIntegrator *integrator, const PSF &psf, const iSGPAR &par)
    : myTheIntegrator(integrator) {
 
    map< string, int > tempAtomList;

    // tempAtomList.insert( <psf.atoms[0].atom_type,1> );
    tempAtomList[psf.atoms[0].atom_type] = 0;
	  
    for(unsigned int i=1;i<psf.atoms.size();i++){
	    
      if( tempAtomList.find( psf.atoms[i].atom_type )
	  == tempAtomList.end() )
	tempAtomList[ psf.atoms[i].atom_type ] = 0;

    }


    //~~~~~~~~~~~~~~~~~~~~~~~
    // Bonds
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all bonds in the par object
    for( unsigned int i = 0; i < par.bonds.size(); i++ ) {

      // determine if this PAR bond is in our system
      if( ( tempAtomList.find( par.bonds[i].atom1 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.bonds[i].atom2 )
	    != tempAtomList.end() ) ) {
	      
        // it is in our system, so store the bond parameters here
        iSGPAR::Bond tempBond = par.bonds[i];
        bonds.push_back(tempBond);
      }
    } // end loop over i

    //~~~~~~~~~~~~~~~~~~~~~~~
    // Angles
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all angles in the par object
    for( unsigned int i = 0; i < par.angles.size(); i++ ) {

      // determine if this PAR angle is in our system
      if( ( tempAtomList.find( par.angles[i].atom1 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.angles[i].atom2 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.angles[i].atom3 )
	    != tempAtomList.end() ) ) {

	// it is in our system, so store the angle parameters here
	iSGPAR::Angle tempAngle = par.angles[i];
	tempAngle.angleval = dtor(par.angles[i].angleval);
	angles.push_back(tempAngle); 
      }   
    } // end loop over i

    //~~~~~~~~~~~~~~~~~~~~~~~
    // Dihedrals
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all dihedrals in the par object
    for( unsigned int i = 0; i < par.dihedrals.size(); i++ ) {

      // determine if this PAR dihedral is in our system
      if( ( tempAtomList.find( par.dihedrals[i].atom1 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.dihedrals[i].atom2 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.dihedrals[i].atom3 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.dihedrals[i].atom4 )
	    != tempAtomList.end() ) ) {

	// it is in our system, so store the angle parameters here
	iSGPAR::Dihedral tempDihedral = par.dihedrals[i];
	for (unsigned int j=0; j < par.dihedrals[i].multiplicity.size(); j++) {
	  for (int k=0; k < par.dihedrals[i].multiplicity[j]; k++)
	    tempDihedral.phaseShift[j][k] = dtor(par.dihedrals[i].phaseShift[j][k]);
	}
	dihedrals.push_back(tempDihedral); 
      }   
    } // end loop over i

    //~~~~~~~~~~~~~~~~~~~~~~~
    // Impropers
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all impropers in the par object
    for( unsigned int i = 0; i < par.impropers.size(); i++ ) {

      // determine if this PAR improper is in our system
      if( ( tempAtomList.find( par.impropers[i].atom1 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.impropers[i].atom2 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.impropers[i].atom3 )
	    != tempAtomList.end() ) &&
	  ( tempAtomList.find( par.impropers[i].atom4 )
	    != tempAtomList.end() ) ) {

	// it is in our system, so store the angle parameters here
	iSGPAR::Improper tempImproper = par.impropers[i];
	for (unsigned int k=0; k < par.impropers[i].phaseShift.size(); k++)
	  tempImproper.phaseShift[k] = dtor(par.impropers[i].phaseShift[k]);	      
	impropers.push_back(tempImproper); 
      }   
    } // end loop over i
  }


  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // doExecute()
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::doExecute(){
    //Report::report << Report::hint << "modifyForces" << Report::endr;
    myTheIntegrator->modifyForces( bonds, angles, dihedrals, impropers, atomTypes, transformMaps );
  }
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // doInitialize()
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::doInitialize() {}
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // indexBonds()
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::indexBonds(GenericTopology *topo, const TRANS &trans) {
    
    //~~~~~~~~~~~~~~~~~~~~~~~
    // Bonds
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all bonds in the topology
    for(unsigned int i=0; i<topo->bonds.size(); i++) {

      // the ID numbers of the bonded atoms
      int I = topo->bonds[i].atom1;
      int J = topo->bonds[i].atom2;

      // the atom type of the atoms in this bond
      string atom1(topo->atomTypes[topo->atoms[I].type].name);
      string atom2(topo->atomTypes[topo->atoms[J].type].name);

      // loop sentinels
      bool found1 = false;
      bool found2 = false;
      unsigned int j;

      // loop over all bonds in the modifier
      for (j=0; j<bonds.size(); j++) {
	
	if (bonds[j].atom1 == atom1 || bonds[j].atom2 == atom1) {found1 = true;}
	if (bonds[j].atom1 == atom2 || bonds[j].atom2 == atom2) {found2 = true;}
	
	if (found1 && found2) {break;}
      } // end loop over j

      topo->bonds[i].iSGmodifierIndex = j;
      
    } // end loop over i

    //~~~~~~~~~~~~~~~~~~~~~~~
    // Angles
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all angles in the topology
    for(unsigned int i=0; i<topo->angles.size(); i++) {
     
      // the ID numbers of the atoms in the angle
      int I = topo->angles[i].atom1;
      int J = topo->angles[i].atom2;
      int K = topo->angles[i].atom3;

      // the atom type of the atoms in this angle
      string atom1 = topo->atomTypes[topo->atoms[I].type].name;
      string atom2 = topo->atomTypes[topo->atoms[J].type].name;
      string atom3 = topo->atomTypes[topo->atoms[K].type].name;

      // loop counter
      unsigned int j;

      // loop over all angles in the modifier
      for (j=0; j<angles.size(); j++) {

	// loop sentinels
	bool found1 = false;
	bool found2 = false;
	bool found3 = false;

	if (angles[j].atom1 == atom1 || angles[j].atom2 == atom1 || angles[j].atom3 == atom1) {found1 = true;}
	if (angles[j].atom1 == atom2 || angles[j].atom2 == atom2 || angles[j].atom3 == atom2) {found2 = true;}
	if (angles[j].atom1 == atom3 || angles[j].atom2 == atom3 || angles[j].atom3 == atom3) {found3 = true;}

	if (found1 && found2 && found3) {break;}
      } // end loop over j
 
      topo->angles[i].iSGmodifierIndex = j;
      
    } // end loop over i

    //~~~~~~~~~~~~~~~~~~~~~~~
    // Dihedrals
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all dihedrals in the topology
    for(unsigned int i=0; i<topo->dihedrals.size(); i++) {
      
      // the ID numbers of the atoms in the dihedral
      int I = topo->dihedrals[i].atom1;
      int J = topo->dihedrals[i].atom2;
      int K = topo->dihedrals[i].atom3;
      int L = topo->dihedrals[i].atom4;

      // the atom type of the atoms in this dihedral
      string atom1 = topo->atomTypes[topo->atoms[I].type].name;
      string atom2 = topo->atomTypes[topo->atoms[J].type].name;
      string atom3 = topo->atomTypes[topo->atoms[K].type].name;
      string atom4 = topo->atomTypes[topo->atoms[L].type].name;

      // loop counter
      unsigned int j;

      // loop over all dihedrals in the modifier
      for (j=0; j<dihedrals.size(); j++) {

	// loop sentinels
	bool found1 = false;
	bool found2 = false;
	bool found3 = false;
	bool found4 = false;
	
	if (dihedrals[j].atom1 == atom1 || dihedrals[j].atom2 == atom1 || dihedrals[j].atom3 == atom1 || dihedrals[j].atom4 == atom1) {found1 = true;}
	if (dihedrals[j].atom1 == atom2 || dihedrals[j].atom2 == atom2 || dihedrals[j].atom3 == atom2 || dihedrals[j].atom4 == atom2) {found2 = true;}
	if (dihedrals[j].atom1 == atom3 || dihedrals[j].atom2 == atom3 || dihedrals[j].atom3 == atom3 || dihedrals[j].atom4 == atom3) {found3 = true;}
	if (dihedrals[j].atom1 == atom4 || dihedrals[j].atom2 == atom4 || dihedrals[j].atom3 == atom4 || dihedrals[j].atom4 == atom4) {found4 = true;}

	if (found1 && found2 && found3 && found4) {break;}
      } // end loop over j
      
      topo->dihedrals[i].iSGmodifierIndex = j;

    } // end loop over dihedrals

    //~~~~~~~~~~~~~~~~~~~~~~~
    // Impropers
    //~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all impropers in the topology
    for(unsigned int i=0; i<topo->impropers.size(); i++) {
      
      // the ID numbers of the atoms in the improper
      int I = topo->impropers[i].atom1;
      int J = topo->impropers[i].atom2;
      int K = topo->impropers[i].atom3;
      int L = topo->impropers[i].atom4;

      // the atom type of the atoms in this improper
      string atom1 = topo->atomTypes[topo->atoms[I].type].name;
      string atom2 = topo->atomTypes[topo->atoms[J].type].name;
      string atom3 = topo->atomTypes[topo->atoms[K].type].name;
      string atom4 = topo->atomTypes[topo->atoms[L].type].name;

      // loop sentinels
      bool found1 = false;
      bool found2 = false;
      bool found3 = false;
      bool found4 = false;
      unsigned int j;

      // loop over all impropers in the modifier
      for (j=0; j<impropers.size(); j++) {
	
	if (impropers[j].atom1 == atom1 || impropers[j].atom2 == atom1 || impropers[j].atom3 == atom1 || impropers[j].atom4 == atom1) {found1 = true;}
	if (impropers[j].atom1 == atom2 || impropers[j].atom2 == atom2 || impropers[j].atom3 == atom2 || impropers[j].atom4 == atom2) {found2 = true;}
	if (impropers[j].atom1 == atom3 || impropers[j].atom2 == atom3 || impropers[j].atom3 == atom3 || impropers[j].atom4 == atom3) {found3 = true;}
	if (impropers[j].atom1 == atom4 || impropers[j].atom2 == atom4 || impropers[j].atom3 == atom4 || impropers[j].atom4 == atom4) {found4 = true;}

	if (found1 && found2 && found3 && found4) {break;}
      } // end loop over j
      
      topo->impropers[i].iSGmodifierIndex = j;

    } // end loop over dihedrals

    //~~~~~~~~~~~~~~~~~~~~~~~
    // AtomTypes
    //~~~~~~~~~~~~~~~~~~~~~~~
    // set the # of transformation stages in the integrator
    myTheIntegrator->myNumStages = trans.NumberOfStages;

    // loop over all atom types in the topology
    for(unsigned int i=0; i<topo->atomTypes.size(); i++) {

      // copy this atom type into the local atomType structure
      atomTypes.push_back(iSGPAR::AtomType(myTheIntegrator->myNumComp,
					   topo->atomTypes[i].name,
					   topo->atomTypes[i].symbolName));

      // copy this atomtype's set of charges for each transformation stage into the ChargeMap structure
      transformMaps.push_back(TransformMap(myTheIntegrator->myNumComp, myTheIntegrator->myNumStages));
    }
      
    // loop over all atom types in the local atomType structure
    for (unsigned int i=0; i<atomTypes.size(); i++) {
      
      // get the name of this atom type
      string myName = atomTypes[i].name;
      
      // loop over all atom types in the mqf file
      for (unsigned int j=0; j<trans.atomTypes.size(); j++) {

	// if this TRANS atomType is the same as our atom type (myName)
	// then store the masses, charges, and alphaLJs for this atom type
	if (myName == trans.atomTypes[j].type_name) {
	  atomTypes[i].mass = trans.atomTypes[j].mass;
          atomTypes[i].charge = trans.atomTypes[j].charge;
          atomTypes[i].stageNumber = trans.atomTypes[j].stage;
          transformMaps[i].old_charge = trans.atomTypes[j].old_charge;
          transformMaps[i].new_charge = trans.atomTypes[j].new_charge;
          transformMaps[i].alphaLJ = trans.atomTypes[j].alphaLJ;
	}

      } // end loop over j
    } // end loop over i

    // fill in the ideal gas state chemical potential difference matrix
    myTheIntegrator->myDeltaMuIG.resize(ArraySizes(myTheIntegrator->myNumComp)
      (myTheIntegrator->myNumComp)(myTheIntegrator->myNumStages));
    for (unsigned int o=0; o<myTheIntegrator->myNumComp; o++)
      for (unsigned int n=0; n<myTheIntegrator->myNumComp; n++) 
        for (unsigned int s=0; s<myTheIntegrator->myNumStages; s++)      
          myTheIntegrator->myDeltaMuIG[o][n][s] = trans.DeltaMuIG[o][n][s];
          
  } // end function indexBonds  

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  checkForTransformation().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::checkForStageCompletion() {
    myTheIntegrator->checkForTransformation(bonds, angles, dihedrals, impropers, atomTypes, transformMaps);}
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  setForcesAfterTransformation().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::setForcesAfterTransformation( GenericTopology *topo ){

    if (myTheIntegrator->Transformed)  {

      unsigned int myT = myTheIntegrator->T;
      unsigned int FinalType = myTheIntegrator->FinalType;
  
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the BOND parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all bonds on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].bondList.begin();
	   iter != topo->molecules[myT].bondList.end(); iter++) {
	
	// find the index # of the proper iSGBond element
	int myISGBond = topo->bonds[*iter].iSGmodifierIndex;
	
	// update the force constant, k
	topo->bonds[*iter].springConstant = bonds[myISGBond].forceConstant[FinalType];    
	
	// update the rest distance, r0
	topo->bonds[*iter].restLength = bonds[myISGBond].distance[FinalType];
	
	// update DeltaK and DeltaR0 
	topo->bonds[*iter].DeltaK = 0.0;
	topo->bonds[*iter].DeltaR0 = 0.0;
      }
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the ANGLE parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all angles on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].angleList.begin();
	   iter != topo->molecules[myT].angleList.end(); iter++) {
	
	// find the index # of the proper iSGAngle element
	int myISGAngle = topo->angles[*iter].iSGmodifierIndex;
	
	// update the force constant, k
	topo->angles[*iter].forceConstant = angles[myISGAngle].forceConstant[FinalType];    
	topo->angles[*iter].ureyBradleyConstant = angles[myISGAngle].k_ub[FinalType];  
	
	// update the rest distance, r0
	topo->angles[*iter].restAngle = angles[myISGAngle].angleval[FinalType];
	topo->angles[*iter].ureyBradleyRestLength = angles[myISGAngle].r_ub[FinalType];  
	
	// update DeltaK and DeltaTheta0
	topo->angles[*iter].DeltaK = 0.0;
	topo->angles[*iter].DeltaTheta0 = 0.0;
	topo->angles[*iter].Delta_ubK = 0.0;
	topo->angles[*iter].Delta_ubR0 = 0.0;
      }
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the DIHEDRAL parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all dihedrals on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].dihedralList.begin();
	   iter != topo->molecules[myT].dihedralList.end(); iter++) {
	
	// find the index # of the proper iSGDihedral element
	int myISGDihedral = topo->dihedrals[*iter].iSGmodifierIndex;
	
	for (int i=0; i<topo->dihedrals[*iter].multiplicity; i++) {
	  // update the force constants
	  topo->dihedrals[*iter].forceConstant[i] = dihedrals[myISGDihedral].forceConstant[FinalType][i];    
	  
	  // update the phase shift angles
	  topo->dihedrals[*iter].phaseShift[i] = dihedrals[myISGDihedral].phaseShift[FinalType][i];
	  
	  // update DeltaK and DeltaPhase
	  topo->dihedrals[*iter].DeltaK[i] = 0.0;
	  topo->dihedrals[*iter].DeltaPhase[i] = 0.0;
	} // end loop over i
      } // end loop over dihedrals
    
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the IMPROPER parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all impropers on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].improperList.begin();
	   iter != topo->molecules[myT].improperList.end(); iter++) {
	
	// find the index # of the proper iSGImproper element
	int myISGImproper = topo->impropers[*iter].iSGmodifierIndex;
	
	// update the force constant 
	topo->impropers[*iter].forceConstant[0] = impropers[myISGImproper].forceConstant[FinalType];    
      
	// update the phase shift angle
	topo->impropers[*iter].phaseShift[0] = impropers[myISGImproper].phaseShift[FinalType];
	
	// update DeltaK and DeltaPhase
	topo->impropers[*iter].DeltaK[0] = 0.0;
	topo->impropers[*iter].DeltaPhase[0] = 0.0;
      }
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the stomic charges and masses
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // the new mass of the molecule
      Real newMolMass = 0.0;

      // loop over all atoms on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].atoms.begin();
	   iter != topo->molecules[myT].atoms.end(); iter++) {

	// the index # into the atomTypes array
	int I = topo->atoms[*iter].type;

	// the atom type name of this atom
	string myName(topo->atomTypes[I].name);

	// find the proper iSGAtomType element
	bool Found = false;
	unsigned int myISGAtomType = 0;
	while (!Found) {
	  if (myName == atomTypes[myISGAtomType].name) Found = true;
	  else myISGAtomType++;
	}

	// update the mass
	topo->atoms[*iter].scaledMass = atomTypes[myISGAtomType].mass[FinalType];
	newMolMass += topo->atoms[*iter].scaledMass;

	// update the charge
	topo->atoms[*iter].scaledCharge = atomTypes[myISGAtomType].charge[FinalType];
	topo->atoms[*iter].scaledCharge *= Constant::SQRTCOULOMBCONSTANT;

	// update deltaM, deltaQ, Qold, and Qnew
	topo->atoms[*iter].deltaM = 0.0;
	topo->atoms[*iter].deltaQ = 0.0;
	topo->atoms[*iter].Qold = 0.0;
	topo->atoms[*iter].Qnew = 0.0;
      } // end loop over atoms

      // update the mass of molecule T
      topo->molecules[myT].mass = newMolMass;
      
    } // end if statement
  } // end function setForcesAfterTransformation()

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  pickNewMolecule().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::pickNewMolecule( GenericTopology *topo, ScalarStructure *myEnergies ) {

    // skip if the molecule has not yet been transformed
    if (myTheIntegrator->Transformed) {

      Real kbT = myTheIntegrator->kbT;
      
      //  randomly select a molecule type to be transformed
      bool Chosen = false;
      while (!Chosen) {

	myTheIntegrator->OldType = (static_cast<int>(randomNumber()*myTheIntegrator->myNumComp)) % myTheIntegrator->myNumComp;

	//  if the # of molecules of the chosen type is zero, then we cannot
	//  choose to transform this type
	if (myTheIntegrator->N[myTheIntegrator->OldType] != 0)
	  Chosen = true;
      }

      unsigned int Old = myTheIntegrator->OldType;
      
      //  choose a random species OldType molecule.
      Chosen = false;
      while (!Chosen) {
      
        // random number
        myTheIntegrator->T = (static_cast<int>(randomNumber()*myTheIntegrator->NumMols)) % myTheIntegrator->NumMols;

        // exit if this molecule type is the same as OldType
        if (topo->molecules[myTheIntegrator->T].type == Old) Chosen = true;
      } // end while loop

      unsigned int myT = myTheIntegrator->T;
      
      // reassign this molecule's Lambda since it is to be transformed
      myTheIntegrator->myLambda = 0.005;
      topo->molecules[myT].lambda = myTheIntegrator->myLambda;
      
      // chose a random species for this molecule to be transformed into
      // this species must not the be same species as myChoice
      Chosen = false;
      while (!Chosen) {
        myTheIntegrator->NewType = (static_cast<int>(randomNumber()*myTheIntegrator->myNumComp)) % myTheIntegrator->myNumComp;
        if (myTheIntegrator->NewType != Old) {Chosen = true;}
      }

      unsigned int New = myTheIntegrator->NewType;
      
      // Update the type and newtype of this molecule
      topo->molecules[myT].type = Old;
      topo->molecules[myT].newtype = New;
          

      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the atomic mass and charge
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      myTheIntegrator->mylnMassRatio = 0.0;
      
      // loop over all atoms on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].atoms.begin();
                                iter != topo->molecules[myT].atoms.end(); iter++) {
        
        // the index # into the atomTypes array
        int I = topo->atoms[*iter].type;
        
        // the atom type name of this atom
        string myName(topo->atomTypes[I].name);
        
        // find the proper iSGAtomType element
        bool Found = false;
        unsigned int myISGAtomType = 0;
        while (!Found) {
          if (myName == atomTypes[myISGAtomType].name) Found = true;
          else myISGAtomType++;
        }       
        
        // update stageNumber, deltaM, deltaQ, Qold, and Qnew, and alphaLJ
        topo->atoms[*iter].stageNumber = atomTypes[myISGAtomType].stageNumber[Old];
        topo->atoms[*iter].Qold = transformMaps[myISGAtomType].old_charge[Old][New][0];
        topo->atoms[*iter].Qnew = transformMaps[myISGAtomType].new_charge[Old][New][0];
        topo->atoms[*iter].deltaQ = transformMaps[myISGAtomType].new_charge[Old][New][0]
          - transformMaps[myISGAtomType].old_charge[Old][New][0];
        topo->atoms[*iter].Qold *= Constant::SQRTCOULOMBCONSTANT;
        topo->atoms[*iter].Qnew *= Constant::SQRTCOULOMBCONSTANT;
        topo->atoms[*iter].deltaQ *= Constant::SQRTCOULOMBCONSTANT;
        topo->atoms[*iter].alphaLJ = transformMaps[myISGAtomType].alphaLJ[Old][New];
        
        // update the following terms only for stage 1 atoms
        if (topo->atoms[*iter].stageNumber == 1) {
          topo->atoms[*iter].deltaM = atomTypes[myISGAtomType].mass[New]
            - atomTypes[myISGAtomType].mass[Old];

          myTheIntegrator->mylnMassRatio += (3 * kbT * 0.5) * ( log(atomTypes[myISGAtomType].mass[Old])
          - log(atomTypes[myISGAtomType].mass[New]) );
        } // end if statement
      } // end loop over atoms
            
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the BOND parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // loop over all bonds on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].bondList.begin();
            iter != topo->molecules[myT].bondList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = topo->atoms[topo->bonds[*iter].atom1].stageNumber;
        int atom2_stage = topo->atoms[topo->bonds[*iter].atom2].stageNumber;

        // determine which stage this bond should be transformed in
        // (it should be the LARGER of the two stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage);

        // update these parameters only if myStage = 1
        if (myStage == 1) {                       	
          // find the index # of the proper iSGBond element
          int myISGBond = topo->bonds[*iter].iSGmodifierIndex;
	
          // update DeltaK and DeltaR0
          topo->bonds[*iter].DeltaK = bonds[myISGBond].forceConstant[New]
            - bonds[myISGBond].forceConstant[Old];
          topo->bonds[*iter].DeltaR0 = bonds[myISGBond].distance[New]
            - bonds[myISGBond].distance[Old];
        }
      }
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the ANGLE parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // loop over all angles on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].angleList.begin();
	   iter != topo->molecules[myT].angleList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = topo->atoms[topo->angles[*iter].atom1].stageNumber;
        int atom2_stage = topo->atoms[topo->angles[*iter].atom2].stageNumber;
        int atom3_stage = topo->atoms[topo->angles[*iter].atom3].stageNumber;
        
        // determine which stage this bond should be transformed in
        // (it should be the LARGEST of the stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage,atom3_stage);

        // update these parameters only if myStage = 1
        if (myStage == 1) {                   
          // find the index # of the proper iSGAngle element
          int myISGAngle = topo->angles[*iter].iSGmodifierIndex;
	
          // update DeltaK and DeltaTheta0
          topo->angles[*iter].DeltaK = angles[myISGAngle].forceConstant[New]
            - angles[myISGAngle].forceConstant[Old];
          topo->angles[*iter].DeltaTheta0 = angles[myISGAngle].angleval[New]
            - angles[myISGAngle].angleval[Old];
          topo->angles[*iter].Delta_ubK = angles[myISGAngle].k_ub[New]
            - angles[myISGAngle].k_ub[Old];
          topo->angles[*iter].Delta_ubR0 = angles[myISGAngle].r_ub[New]
            - angles[myISGAngle].r_ub[Old];
        }
      }
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the DIHEDRAL parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // loop over all dihedrals on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].dihedralList.begin();
	   iter != topo->molecules[myT].dihedralList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = topo->atoms[topo->dihedrals[*iter].atom1].stageNumber;
        int atom2_stage = topo->atoms[topo->dihedrals[*iter].atom2].stageNumber;
        int atom3_stage = topo->atoms[topo->dihedrals[*iter].atom3].stageNumber;
        int atom4_stage = topo->atoms[topo->dihedrals[*iter].atom4].stageNumber;
        
        // determine which stage this bond should be transformed in
        // (it should be the LARGEST of the stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);

        // update these parameters only if myStage = 1
        if (myStage == 1) {                   
          // find the index # of the proper iSGDihedral element
          int myISGDihedral = topo->dihedrals[*iter].iSGmodifierIndex;
	
          for (int i=0; i<topo->dihedrals[*iter].multiplicity; i++) {
            
            // update DeltaK and DeltaPhase
            topo->dihedrals[*iter].DeltaK[i] = dihedrals[myISGDihedral].forceConstant[New][i]
              - dihedrals[myISGDihedral].forceConstant[Old][i];
            topo->dihedrals[*iter].DeltaPhase[i] = dihedrals[myISGDihedral].phaseShift[New][i]
              - dihedrals[myISGDihedral].phaseShift[Old][i];
          } // end loop over i
        } // end if statement
      } // end loop over dihedrals
        
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the IMPROPER parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // loop over all impropers on molecule T
      for (std::vector<int>::iterator iter = topo->molecules[myT].improperList.begin();
	   iter != topo->molecules[myT].improperList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = topo->atoms[topo->impropers[*iter].atom1].stageNumber;
        int atom2_stage = topo->atoms[topo->impropers[*iter].atom2].stageNumber;
        int atom3_stage = topo->atoms[topo->impropers[*iter].atom3].stageNumber;
        int atom4_stage = topo->atoms[topo->impropers[*iter].atom4].stageNumber;
        
        // determine which stage this bond should be transformed in
        // (it should be the LARGEST of the stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);

        // update these parameters only if myStage = 1
        if (myStage == 1) {     	
          // find the index # of the proper iSGImproper element
          int myISGImproper = topo->impropers[*iter].iSGmodifierIndex;
	
          // update DeltaK and DeltaPhase
          topo->impropers[*iter].DeltaK[0] = impropers[myISGImproper].forceConstant[New]
            - impropers[myISGImproper].forceConstant[Old];
          topo->impropers[*iter].DeltaPhase[0] = impropers[myISGImproper].phaseShift[New]
            - impropers[myISGImproper].phaseShift[Old];
        }
      }

      // Compute the residual and ideal gas chemical potential differences:
      // ideal gas chemical potential difference (using the precomputed chemical potentials)
      myTheIntegrator->myDMuIG = myTheIntegrator->myDeltaMuIG[Old][New][0];

      // residual chemical potential difference
      // DeltaMu = kT*ln(f1/f0 * N0/(N1+1))
      myTheIntegrator->myTargetDeltaMu = myTheIntegrator->myTargetMu[New] - myTheIntegrator->myTargetMu[Old]
	+ kbT * log (static_cast<Real>(myTheIntegrator->N[Old]) / static_cast<Real>(myTheIntegrator->N[New]+1));
      myTheIntegrator->myTargetDeltaMu /= static_cast<Real>(myTheIntegrator->myNumStages);
      
      // new Lambda velocity chosen from MB distribution
      // width of Maxwell-Boltzmann distribution (in units of [1/fs])
      Real sigma = sqrt(kbT / myTheIntegrator->Qd);
      
      // assign a new positive lambda velocity
      Real width = randomGaussianNumber(0.0,sigma);
      width *= width;
      myTheIntegrator->myLambdaVel = sqrt(width);
      
      //  reset the conserved quantity averages
      myTheIntegrator->AveCQ = 0.0;
      myTheIntegrator->AveCQSq = 0.0;
      myTheIntegrator->NumTransSteps = 0;
      
      // reset the transformation flag
      myTheIntegrator->Transformed = false;
      
      // reset AllEnergiesFile flag
      myEnergies->output(false);
      (*myEnergies).clear();
    } // end if statement
  } // end function pickNewMolecule()

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // readXSCs (eXstended System Coordinates)
  // read this only if an XSC file is specified in the config file.
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void ModifierISG::readXSCs(const string myFile, GenericTopology *topo) {

    // create the XSC reader object
    XSCReader xscReader;

    if(!xscReader.open(myFile))
      report << error << "Can't open XSC file \'"<< myFile <<"\'."<<endr;

    // create the storage object for the XSC information
    XSC xsc;

    // read in the XSC file, which will contain the integrator's
    // time-zero extended system variable values
    if(!(xscReader >> xsc))
      report << error << "Could not parse XSC file \'"<< myFile <<"\'."<<endr;
    report << plain << "Using XSC file \'"<< myFile <<"\'."<< endr;

    // store the time-zero values of the extended system variables
    myTheIntegrator->T = xsc.myMolecule-1;
    myTheIntegrator->myLambda = xsc.Lambda;
    myTheIntegrator->myLambdaVel = xsc.Lambda_vel;
    myTheIntegrator->OldType = xsc.old_type;
    myTheIntegrator->NewType = xsc.new_type;
    myTheIntegrator->myEta = xsc.Eta;
    myTheIntegrator->myEtaV = xsc.EtaVol;
    myTheIntegrator->myEtaVel = xsc.Eta_vel;
    myTheIntegrator->myEtaVolVel = xsc.EtaVol_vel;
    myTheIntegrator->myVolume = xsc.Vol;
    myTheIntegrator->myEpsilonVel = xsc.Epsilon_vel;

    myTheIntegrator->Transformed = false;

    report << "Initial values of the extended system variables:" << endr;
    report << "Transforming molecule: " << myTheIntegrator->T+1 << endr;
    report << "OldType = " << myTheIntegrator->OldType << endr;
    report << "NewType = " << myTheIntegrator->NewType << endr;
    report.precision(10);
    report << "Lambda = " << myTheIntegrator->myLambda << endr;
    report << "Lambda velocity (1/fs) = " << myTheIntegrator->myLambdaVel << endr;
    report << "Eta = " << myTheIntegrator->myEta << endr;
    report << "Eta velocity (1/fs) = " << myTheIntegrator->myEtaVel << endr;
    report << "EtaVol = " << myTheIntegrator->myEtaV << endr;
    report << "EtaVol velocity (1/fs) = " << myTheIntegrator->myEtaVolVel << endr;
    report << "Volume = " << myTheIntegrator->myVolume << endr;
    report << "Epsilon velocity (1/fs) = " << myTheIntegrator->myEpsilonVel << endr;

    // compute the ideal gas chemical potential
    // difference from changing the molecule's mass
    myTheIntegrator->mylnMassRatio = 0.0;
    int myT = myTheIntegrator->T;
    int Old = myTheIntegrator->OldType;
    int New = myTheIntegrator->NewType;
    Real kbT = myTheIntegrator->kbT;

    // determine which transformation state the molecule is currently in
    myTheIntegrator->thisStage = static_cast<int>(floor(xsc.Lambda)) + 1;

    // loop over all atoms on molecule T
    for (std::vector<int>::iterator iter = topo->molecules[myT].atoms.begin();
	 iter != topo->molecules[myT].atoms.end(); iter++) {

      // the index # into the atomTypes array
      int I = topo->atoms[*iter].type;

      // the atom type name of this atom
      string myName(topo->atomTypes[I].name);

      // find the proper iSGAtomType element
      bool Found = false;
      unsigned int myISGAtomType = 0;
      while (!Found) {
        if (myName == atomTypes[myISGAtomType].name) Found = true;
        else myISGAtomType++;
      }

      // update stageNumber, deltaM, deltaQ, Qold, and Qnew, and alphaLJ
      topo->atoms[*iter].stageNumber = atomTypes[myISGAtomType].stageNumber[Old];
      topo->atoms[*iter].Qold = transformMaps[myISGAtomType].old_charge[Old][New][myTheIntegrator->thisStage-1];
      topo->atoms[*iter].Qnew = transformMaps[myISGAtomType].new_charge[Old][New][myTheIntegrator->thisStage-1];
      topo->atoms[*iter].deltaQ = transformMaps[myISGAtomType].new_charge[Old][New][myTheIntegrator->thisStage-1]
        - transformMaps[myISGAtomType].old_charge[Old][New][myTheIntegrator->thisStage-1];
      topo->atoms[*iter].Qold *= Constant::SQRTCOULOMBCONSTANT;
      topo->atoms[*iter].Qnew *= Constant::SQRTCOULOMBCONSTANT;
      topo->atoms[*iter].deltaQ *= Constant::SQRTCOULOMBCONSTANT;
      topo->atoms[*iter].alphaLJ = transformMaps[myISGAtomType].alphaLJ[Old][New];
      // add the following terms only for the current transformation stage
      if (topo->atoms[*iter].stageNumber == myTheIntegrator->thisStage) {
        topo->atoms[*iter].deltaM = atomTypes[myISGAtomType].mass[New]
          - atomTypes[myISGAtomType].mass[Old];
        
        // update mylnMassRatio for the current stage
        myTheIntegrator->mylnMassRatio += (3 * kbT * 0.5) * ( log(atomTypes[myISGAtomType].mass[Old]) - log(atomTypes[myISGAtomType].mass[New]) );
      } // end if statement
    } // end loop over atoms

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the BOND parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all bonds on molecule T
    for (std::vector<int>::iterator iter = topo->molecules[myT].bondList.begin();
         iter != topo->molecules[myT].bondList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = topo->atoms[topo->bonds[*iter].atom1].stageNumber;
      int atom2_stage = topo->atoms[topo->bonds[*iter].atom2].stageNumber;

      // determine which stage this bond should be transformed in
      // (it should be the LARGER of the two stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage);

      // update these parameters only if myStage = thisStage
      if (myStage == myTheIntegrator->thisStage) {
        // find the index # of the proper iSGBond element
        int myISGBond = topo->bonds[*iter].iSGmodifierIndex;
        
        // update DeltaK and DeltaR0
        topo->bonds[*iter].DeltaK = bonds[myISGBond].forceConstant[New]
          - bonds[myISGBond].forceConstant[Old];
        topo->bonds[*iter].DeltaR0 = bonds[myISGBond].distance[New]
          - bonds[myISGBond].distance[Old];
      }
    }
      
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the ANGLE parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all angles on molecule T
    for (std::vector<int>::iterator iter = topo->molecules[myT].angleList.begin();
         iter != topo->molecules[myT].angleList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = topo->atoms[topo->angles[*iter].atom1].stageNumber;
      int atom2_stage = topo->atoms[topo->angles[*iter].atom2].stageNumber;
      int atom3_stage = topo->atoms[topo->angles[*iter].atom3].stageNumber;
        
      // determine which stage this bond should be transformed in
      // (it should be the LARGEST of the stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage,atom3_stage);
       
      // update these parameters only if myStage = thisStage
      if (myStage == myTheIntegrator->thisStage) {
        // find the index # of the proper iSGAngle element
        int myISGAngle = topo->angles[*iter].iSGmodifierIndex;
        
        // update DeltaK and DeltaTheta0
        topo->angles[*iter].DeltaK = angles[myISGAngle].forceConstant[New]
          - angles[myISGAngle].forceConstant[Old];
        topo->angles[*iter].DeltaTheta0 = angles[myISGAngle].angleval[New]
          - angles[myISGAngle].angleval[Old];
        topo->angles[*iter].Delta_ubK = angles[myISGAngle].k_ub[New]
          - angles[myISGAngle].k_ub[Old];
        topo->angles[*iter].Delta_ubR0 = angles[myISGAngle].r_ub[New]
          - angles[myISGAngle].r_ub[Old];
      }
    }
      
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the DIHEDRAL parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all dihedrals on molecule T
    for (std::vector<int>::iterator iter = topo->molecules[myT].dihedralList.begin();
         iter != topo->molecules[myT].dihedralList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = topo->atoms[topo->dihedrals[*iter].atom1].stageNumber;
      int atom2_stage = topo->atoms[topo->dihedrals[*iter].atom2].stageNumber;
      int atom3_stage = topo->atoms[topo->dihedrals[*iter].atom3].stageNumber;
      int atom4_stage = topo->atoms[topo->dihedrals[*iter].atom4].stageNumber;
        
      // determine which stage this bond should be transformed in
      // (it should be the LARGEST of the stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);
       
      // update these parameters only if myStage = thisStage
      if (myStage == myTheIntegrator->thisStage) {
        // find the index # of the proper iSGDihedral element
        int myISGDihedral = topo->dihedrals[*iter].iSGmodifierIndex;
        
        for (int i=0; i<topo->dihedrals[*iter].multiplicity; i++) {            
          // update DeltaK and DeltaPhase
          topo->dihedrals[*iter].DeltaK[i] = dihedrals[myISGDihedral].forceConstant[New][i]
            - dihedrals[myISGDihedral].forceConstant[Old][i];
          topo->dihedrals[*iter].DeltaPhase[i] = dihedrals[myISGDihedral].phaseShift[New][i]
            - dihedrals[myISGDihedral].phaseShift[Old][i];
        } // end loop over i
      } // end if statement
    } // end loop over dihedrals
        
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the IMPROPER parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all impropers on molecule T
    for (std::vector<int>::iterator iter = topo->molecules[myT].improperList.begin();
         iter != topo->molecules[myT].improperList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = topo->atoms[topo->impropers[*iter].atom1].stageNumber;
      int atom2_stage = topo->atoms[topo->impropers[*iter].atom2].stageNumber;
      int atom3_stage = topo->atoms[topo->impropers[*iter].atom3].stageNumber;
      int atom4_stage = topo->atoms[topo->impropers[*iter].atom4].stageNumber;
        
      // determine which stage this bond should be transformed in
      // (it should be the LARGEST of the stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);

      // update these parameters only if myStage = thisStage
      if (myStage == myTheIntegrator->thisStage) {
        // find the index # of the proper iSGImproper element
        int myISGImproper = topo->impropers[*iter].iSGmodifierIndex;
        
        // update DeltaK and DeltaPhase
        topo->impropers[*iter].DeltaK[0] = impropers[myISGImproper].forceConstant[New]
          - impropers[myISGImproper].forceConstant[Old];
        topo->impropers[*iter].DeltaPhase[0] = impropers[myISGImproper].phaseShift[New]
          - impropers[myISGImproper].phaseShift[Old];
      }
    } // end loop over impropers
  } // end function readXSCs()
}  // end class ModifierISG

