//  -----------------------------------------------------------------------  //
//  explicit, time-reversible integrator for iSG dynamics                    //
//                                                                           //
//  Unless modified, this integrator uses the molecular virial to control    //
//  the pressure. -- TIM                                                     //
//  -----------------------------------------------------------------------  //

#include "iSGIntegrator.h"
#include "ScalarStructure.h" 
#include "Vector3DBlock.h"
#include "ForceGroup.h"
#include "GenericTopology.h" 
#include "pmconstants.h"
#include "topologyutilities.h"
#include "ModifierPostForceBarostat.h"
#include "iSGModifierPostForceChemostat.h"
#include "ModifierPostForceThermostat.h"
#include "ModifierPreForceBarostat.h"
#include "iSGModifierPreForceChemostat.h"
#include "ModifierPreForceThermostat.h"
#include "ModifierNPTShake.h"
#include "ModifierNPTRattle.h"

//#define DEBUG_MODIFIER_BOND
//#define DEBUG_MODIFIER_ANGLE
//#define DEBUG_MODIFIER_TORSION
//#define DEBUG_MODIFIER_ATOM
//#define DEBUG_CHEMOSTAT

using std::vector;
using std::string;
using namespace ProtoMol::Report;

namespace ProtoMol {

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Keyword.
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  const string iSGIntegrator::keyword("iSGVerlet");

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Default or empty constructor
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  iSGIntegrator::iSGIntegrator(): STSIntegrator(),
				  myNumComp(0),
                                  myNumStages(0),
				  myTargetTemp(0.0), 
				  myTargetPres(0.0),
				  myTauT(0.0), 
				  myTauV(0.0), 
				  myTauP(0.0),
				  myTauD(0.0),
                                  kbT(0.0),
                                  myLambda(0.0),
                                  NumAtoms(0),
                                  NumMols(0),
                                  myNumFree(0),
				  T(-1),
                                  thisStage(0),
				  NumTransSteps(0),
				  OldType(0), 
				  NewType(0),
				  Transformed(true),
				  myTargetDeltaMu(0.0),
                                  mylnMassRatio(0.0),
                                  myDMuIG(0.0),
				  myVolume(0.0),
				  myEpsilonVel(0.0),
				  Qo(0.0),
				  Qv(0.0),
				  W(0.0),
				  Qd(0.0),
				  myEta(0.0),
				  myEtaV(0.0),
				  myEtaVel(0.0),
				  myEtaVolVel(0.0),
				  myLambdaVel(0.0),
				  AveCQ(0.0),
				  AveCQSq(0.0),
                                  AveDeltaMu(0.0) {}

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Constructor
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  iSGIntegrator::iSGIntegrator(Real timestep, 
                               unsigned int numComp,
			       Real temperature,
                               Real pressure, 
                               const vector<Real> &fugacityFrac,
                               Real tauT, 
			       Real tauV,
                               Real tauP, 
			       Real tauD,
                               ForceGroup *overloadedForces) 
    : STSIntegrator(timestep, overloadedForces),
      myNumComp(numComp),
      myNumStages(0),
      myTargetTemp(temperature), 
      myTargetPres(pressure),
      myTauT(tauT), 
      myTauV(tauV), 
      myTauP(tauP),
      myTauD(tauD),
      kbT(temperature * Constant::BOLTZMANN),
      myLambda(0.0),
      NumAtoms(0),
      NumMols(0),
      myNumFree(0),
      T(-1),
      thisStage(1),
      NumTransSteps(0),
      OldType(0), 
      NewType(0), 
      Transformed(true),
      myTargetDeltaMu(0.0),
      mylnMassRatio(0.0),
      myDMuIG(0.0),
      myVolume(0.0),
      myEpsilonVel(0.0),
      Qo(0.0),
      Qv(0.0),
      W(0.0),
      Qd(0.0),
      myEta(0.0),
      myEtaV(0.0),
      myEtaVel(0.0),
      myEtaVolVel(0.0),
      myLambdaVel(0.0),
      AveCQ(0.0),
      AveCQSq(0.0),
      AveDeltaMu(0.0) {

    if(numComp != fugacityFrac.size())
      report << error << "[iSGIntegrator::iSGIntegrator] oh no, you tried to pass fugacityFrac("<<fugacityFrac.size()<<") with a different size than number of components("<<numComp<<")... bye!"<<endr;

    // store the fugacity fractions --> kT * ln (fugacityFrac)
    myTargetMu.resize(myNumComp,0.0);
    myFugacityFrac = fugacityFrac;

    Real sum = 0.0;

    for (unsigned int i=0; i<myNumComp; i++) {
      myTargetMu[i] = myTargetTemp * Constant::BOLTZMANN * log(fugacityFrac[i]);
      sum += fugacityFrac[i];
    }

    // Stop if the fugacity fractions do not sum to 1
    if (!equal(sum,1.0,1e-13))
      report << error << "Fugacity fractions do not sum (=1 + "<<sum-1.0<<") to 1."<< endr;   

  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Thermostat -- prior to force calculations
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::PreForceThermostat() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();   
    //  Twice the kinetic energy.  Units: (kcal / mol)
    const Real twiceKE = 2. * kineticEnergy(myTopo, myVelocities);


    //  Advance the particle thermostat variable.
    myEta += myEtaVel * halfDeltaT;
    //  Advance the particle thermostat variable velocity.
    myEtaVel += (twiceKE - myNumFree * kbT) * halfDeltaT / Qo;
  
  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Thermostat -- after force calculations
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::PostForceThermostat() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();    
    //  Get the new KE.
    const Real twiceKE = 2. * kineticEnergy(myTopo, myVelocities);


    //  Advance the particle thermostat variable velocity.
    myEtaVel += (twiceKE - myNumFree * kbT) * halfDeltaT / Qo;
    //  Advance the particle thermostat variable.
    myEta += myEtaVel * halfDeltaT;
   
    //  New particle thermostat kinetic energy.
    Real pEta = (Qo * 0.5) * (myEtaVel * myEtaVel);
    //  New particle thermostat potential energy.
    Real VEta = myEta * myNumFree * kbT;

    //  Add the energy from the extended system thermostat.
    (*myEnergies)[ScalarStructure::INTEGRATOR] += pEta + VEta;

  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Barostat -- prior to force calculations
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::PreForceBarostat() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();    
    //  Twice the kinetic energy.  Units: (kcal / mol)
    const Real twiceKE = 2. * kineticEnergy(myTopo, myVelocities);
    const Real MolKE = molecularKineticEnergy(myTopo, myVelocities);


    //  Advance the volume thermostat variable.
    myEtaV += myEtaVolVel * halfDeltaT;
    //  Advance the volume thermostat variable velocity.
    myEtaVolVel += (W * myEpsilonVel * myEpsilonVel - kbT) * halfDeltaT / Qv;
    
    //  Calculate the current molecular pressure.  Units: (bar)
    Real currentPres = computeMolecularPressure(myEnergies, myVolume, MolKE);

    //  Advance the box volume velocity.  Divide by PRESSUREFACTOR to get
    //  correct (fs)^-1 units.
    myEpsilonVel += (3.0 * (myVolume) * (currentPres - myTargetPres)) *
      halfDeltaT / (Constant::PRESSUREFACTOR * W);
    //  Advance the box volume velocity due to thermostat forces.
    myEpsilonVel += (3.0 * twiceKE / myNumFree) * halfDeltaT / W;
    myEpsilonVel *= exp(-myEtaVolVel * halfDeltaT);

    //  Advance the box volume.
    Real fac = exp(3.0 * myEpsilonVel * 2. * halfDeltaT);
    myVolume *= fac;
    myTopo->rescaleVolume(fac);
    uncache();

  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Barostat -- after force calculations
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::PostForceBarostat() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();   
    //  Get the new KE.
    const Real twiceKE = 2. * kineticEnergy(myTopo, myVelocities);
    const Real MolKE = molecularKineticEnergy(myTopo, myVelocities);

    //  Calculate the current pressure.  Units: (bar)
    Real currentPres = computeMolecularPressure(myEnergies, myVolume, MolKE);

    //  Advance the box volume velocity.
    myEpsilonVel *= exp(-myEtaVolVel * halfDeltaT);
    myEpsilonVel += (3.0 * twiceKE / myNumFree) * halfDeltaT / W;
    //  Advance the box volume velocity.  Divide by PRESSUREFACTOR to get
    //  correct (fs)^-1 units.
    myEpsilonVel += (3.0 * (myVolume) * (currentPres - myTargetPres)) *
      halfDeltaT / (Constant::PRESSUREFACTOR * W);

    //  Advance the volume thermostat variable velocity.
    myEtaVolVel += (W * myEpsilonVel * myEpsilonVel - kbT) * halfDeltaT / Qv;
    //  Advance the volume thermostat variable.
    myEtaV += myEtaVolVel * halfDeltaT;


    //  New volume thermostat kinetic energy.
    Real pEtaV = (Qv * 0.5) * (myEtaVolVel * myEtaVolVel);
    //  New volume thermostat potential energy
    Real VEtaV = myEtaV * kbT;
    //  New box volume kinetic energy.
    Real pVol = (W * 0.5) * (myEpsilonVel * myEpsilonVel);
    //  New box volume potential energy.
    Real VVol = (myVolume * myTargetPres) / Constant::PRESSUREFACTOR;
    
    //  Add the energy from the extended system barostat.
    (*myEnergies)[ScalarStructure::INTEGRATOR] += pEtaV + VEtaV + pVol + VVol;

  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Chemostat -- prior to force calculations
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::PreForceChemostat() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();   


    //  Advance the chemostat velocity
    Real deltaM_vsq = 0.0;
    for (unsigned int i=0; i < myTopo->molecules[T].size(); i++) {   
      //  Current atom # and mass
      int atom = myTopo->molecules[T][i];
      deltaM_vsq += myTopo->atoms[atom].deltaM * ((*myVelocities)[atom]).normSquared();
    }

    myLambdaVel += (myTargetDeltaMu + myDMuIG - (*myEnergies).deltaMu()
      + (deltaM_vsq * 0.5) + mylnMassRatio) * halfDeltaT / Qd;

    //  Advance the chemostat
    myLambda += myLambdaVel * 2. * halfDeltaT;

    //  If Lambda should become negative, reset it to be zero,
    //  and if Lambda should become larger than myNumStages, reset it to be myNumStages.
    if (myLambda < 0) myLambda = 0.0;
    else if (myLambda > myNumStages) myLambda = static_cast<Real>(myNumStages);

  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Chemostat -- after force calculations
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::PostForceChemostat() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();    

    
    //  Advance the chemostat velocity
    Real deltaM_vsq = 0.0;
    for (unsigned int i=0; i < myTopo->molecules[T].size(); i++) {  
      //  Current atom # and mass
      int atom = myTopo->molecules[T][i];
      deltaM_vsq += myTopo->atoms[atom].deltaM * ((*myVelocities)[atom]).normSquared();
    }

    myLambdaVel += (myTargetDeltaMu + myDMuIG - (*myEnergies).deltaMu()
      + (deltaM_vsq * 0.5) + mylnMassRatio ) * halfDeltaT / Qd;
           
#ifdef DEBUG_CHEMOSTAT
    report.precision(6);
    report << hint << "*******************************************************" << endr;
    report << hint << "Molecule being transformed: " << T+1 << ", current step: " << NumTransSteps << endr;
    report << hint << "Current identity is: " << OldType << ", New identity is: " << NewType << endr;
    report << hint << "deltaM_vsq = " << deltaM_vsq << endr;
    report << hint << "TargetDeltaMu = " << myTargetDeltaMu << ", mass ratio term = "
                   << mylnMassRatio << ", DeltaMuIG = " << myDMuIG << endr;
    report << hint << "Lambda = " << myLambda << endr;
    report << hint << "Drift Current DeltaMu = " << (*myEnergies).deltaMu() << endr;
    report << hint << "Bond DeltaMu = " << (*myEnergies)[ScalarStructure::BOND_DELTAMU] << endr;
    report << hint << "Angle DeltaMu = " << (*myEnergies)[ScalarStructure::ANGLE_DELTAMU] << endr;
    report << hint << "Dihedral DeltaMu = " << (*myEnergies)[ScalarStructure::DIHEDRAL_DELTAMU] << endr;
    report << hint << "Improper DeltaMu = " << (*myEnergies)[ScalarStructure::IMPROPER_DELTAMU] << endr;
    report << hint << "LennardJones DeltaMu = " << (*myEnergies)[ScalarStructure::LENNARDJONES_DELTAMU] << endr;
    report << hint << "Coulomb DeltaMu = " << (*myEnergies)[ScalarStructure::COULOMB_DELTAMU] << endr;
    report << hint << "LambdaVel = " << myLambdaVel << endr;
    report << hint << "*******************************************************" << endr;
#endif
         
    //  New chemostat kinetic energy
    Real pLambda = (Qd * 0.5) * (myLambdaVel * myLambdaVel);
    //  New chemostat potential energy
    //Real VLambda = (thisStage - 1 - myLambda) * (myTargetDeltaMu + mylnMassRatio);
    Real VLambda = -myLambda * myTargetDeltaMu
     + (thisStage - 1 - myLambda) * (myDMuIG + mylnMassRatio);
        
    //  Add the energy from the extended system chemostat.
    (*myEnergies)[ScalarStructure::INTEGRATOR] += pLambda + VLambda;
    //  add to the average chemical potential difference
    AveDeltaMu += (*myEnergies).deltaMu();

    //  add to the number of timesteps used to transform this molecule
    NumTransSteps++;
  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  doHalfKick().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::doHalfKick() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();

    // calculate the COM and momentum of each molecule
    buildMolecularCenterOfMass(myPositions,myTopo);
    buildMolecularMomentum(myVelocities,myTopo);

    // ---------------------------------------------------------------------
    //  Do the first update of the atom velocities
    //  Loop over all molecules
    for (unsigned int i=0; i < NumMols; i++) {
      //  Temporary storage element for the updated molecular momentum
      Vector3D Momentum(0.0,0.0,0.0);

      //  Loop over the atoms on this molecule
      for (unsigned int a=0; a < myTopo->molecules[i].size(); a++) {

        //  Current atom # and mass
        int atom = myTopo->molecules[i][a];
        Real mass = myTopo->atoms[atom].scaledMass;

        //  Advance the velocities due to the thermostat and barostat forces.
        (*myVelocities)[atom] *= exp(-(mass / myTopo->molecules[i].mass * (1.0 + 3.0/myNumFree)
				       * myEpsilonVel + myEtaVel ) * halfDeltaT);

        //  Add to the new momentum of this molecule
        Momentum += (*myVelocities)[atom] * mass;

      } //  end loop over atoms
      
      //  Store the updated molecular momentum
      myTopo->molecules[i].momentum = Momentum;
    } //  end loop over molecules

    // -------------------------------------------------------------------------
    //  Do the second and third updates of the atom velocities
    //  Loop over all molecules
    for (unsigned int i=0; i < NumMols; i++) {

      //  Loop over the atoms on this molecule
      for (unsigned int a=0; a < myTopo->molecules[i].size(); a++) {
  
        //  Current atom # and mass
        int atom = myTopo->molecules[i][a];
        Real mass = myTopo->atoms[atom].scaledMass;

        //  Advance the velocities due to barostat force.
        (*myVelocities)[atom] -= (myTopo->molecules[i].momentum - (*myVelocities)[atom] * mass)
	  * (1.0 + 3.0/myNumFree) * (1.0 / myTopo->molecules[i].mass)
	  * myEpsilonVel * halfDeltaT;

        //  Advance the velocities due to atomic forces.
        (*myVelocities)[atom] += (*myForces)[atom] * halfDeltaT * Constant::INV_TIMEFACTOR / mass;
      }  // end loop over atoms
    } //  end loop over molecules
  }  //  End doHalfKick().

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  doDrift().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
  void iSGIntegrator::doDrift() {

    //  Timestep.  Units: (fs)
    const Real deltaT = getTimestep();

    // -------------------------------------------------------------------------
    //  Do the first update of the atom positions
    //  Loop over all molecules
    for (unsigned int i=0; i < NumMols; i++) {

      //  Loop over the atoms on this molecule
      for (unsigned int a=0; a < myTopo->molecules[i].size(); a++) {
  
        //  Current atom # and mass
        int atom = myTopo->molecules[i][a];
        Real mass = myTopo->atoms[atom].scaledMass;

        //  Advance the positions due to change in box length.
        (*myPositions)[atom] *= exp((mass / myTopo->molecules[i].mass)
				    * myEpsilonVel * deltaT);
      }  // end loop over atoms
    } //  end loop over molecules

    // -------------------------------------------------------------------------
    //  Do the second and final update of the positions
    //  Loop over all molecules
    for (unsigned int i=0; i < NumMols; i++) {

      //  Temporary storage element for the updated molecular COM
      Vector3D COM(0.0,0.0,0.0);

      //  Loop over the atoms on this molecule
      for (unsigned int a=0; a < myTopo->molecules[i].size(); a++) {
  
        //  Current atom # and mass
        int atom = myTopo->molecules[i][a];
        Real mass = myTopo->atoms[atom].scaledMass;

        //  Advance the positions due to box volume or barostat (similar to COM shifting).
        (*myPositions)[atom] += (myTopo->molecules[i].position - (*myPositions)[atom] 
				 * (mass / myTopo->molecules[i].mass)) * deltaT * myEpsilonVel;

        //  Advance the positions due to velocity.
        (*myPositions)[atom] += (*myVelocities)[atom] * deltaT * Constant::INV_TIMEFACTOR;

        //  Add to the new COM of this molecule
        COM += (*myPositions)[atom] * mass;
      }  //  end loop over atoms
      
      //  Store the updated molecular COM
      myTopo->molecules[i].position = COM / (myTopo->molecules[i].mass);
    } //  end loop over molecules

    //  update the COM of each molecule
    buildMolecularCenterOfMass(myPositions,myTopo);
  }  // end doDrift()
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  do2ndHalfkick()
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::do2ndHalfKick() {

    //  Timestep.  Units: (fs)
    const Real halfDeltaT = 0.5 * getTimestep();

    // -----------------------------------------------------------------------------
    //  Do the first update of the atomic velocities
    //  Loop over all molecules
    for (unsigned int i=0; i < NumMols; i++) {

      //  Temporary storage element for the updated molecular momentum
      Vector3D Momentum(0.0,0.0,0.0);

      //  Loop over the atoms on this molecule
      for (unsigned int a=0; a < myTopo->molecules[i].size(); a++) {
  
        //  Current atom # and mass
        int atom = myTopo->molecules[i][a];
        Real mass = myTopo->atoms[atom].scaledMass;

        //  Advance the velocities due to atomic forces.
        (*myVelocities)[atom] += (*myForces)[atom] * halfDeltaT * Constant::INV_TIMEFACTOR / mass;

        //  Add to the new momentum of this molecule
        Momentum += (*myVelocities)[atom] * mass;
      }  //  end loop over atoms

      //  Store the updated molecular momentum
      myTopo->molecules[i].momentum = Momentum;
    } //  end loop over molecules

    // ------------------------------------------------------------------------------
    //  Do the second and third updates of the atomic velocities
    //  Loop over all molecules
    for (unsigned int i=0; i < NumMols; i++) {
       
      //  Temporary storage element for the updated molecular momentum
      Vector3D Momentum(0.0,0.0,0.0);

      //  Loop over the atoms on this molecule
      for (unsigned int a=0; a < myTopo->molecules[i].size(); a++) {
  
        //  Current atom # and mass
        int atom = myTopo->molecules[i][a];
        Real mass = myTopo->atoms[atom].scaledMass;

        //  Advance the velocities due to barostat force.
        (*myVelocities)[atom] -= (myTopo->molecules[i].momentum - (*myVelocities)[atom] * mass)
	  * (1.0 + 3.0/myNumFree) * (1.0 / myTopo->molecules[i].mass)
	  * myEpsilonVel * halfDeltaT;

        //  Advance the velocities due to the thermostat and barostat forces.
        (*myVelocities)[atom] *= exp(-(mass / myTopo->molecules[i].mass * (1.0 + 3.0/myNumFree)
          * myEpsilonVel + myEtaVel ) * halfDeltaT);

        //  Add to the new momentum of this molecule
        Momentum += (*myVelocities)[atom] * mass;                                 
      }  // end loop over atoms

      //  Store the updated molecular momentum
      myTopo->molecules[i].momentum = Momentum;
    } //  end loop over molecules
  }  // end do2ndHalfKick
      
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  modifyForces().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::modifyForces(const vector< iSGPAR::Bond > &myBonds,
				   const vector< iSGPAR::Angle > &myAngles,
				   const vector< iSGPAR::Dihedral > &myDihedrals,
				   const vector< iSGPAR::Improper > &myImpropers,
				   const vector< iSGPAR::AtomType > &myAtomTypes,
                                   const vector< ModifierISG::TransformMap > &myTransformMaps ){

    // exit if we have not chosen a transforming molecule yet
    if (T == -1) return;

    // if lambda should be larger than the current stage # or smaller than the previous stage # then
    // use a temporary lambda = newStage or lambda = oldStage so that we don't end up multiplying any force
    // parameters by a number that is larger than 1 or less than 0.
    int oldStage = thisStage - 1;
    Real PreModLambda = myLambda;

    // if myLambda is larger than thisStage then temporarily set myLambda to be just less than thisStage;
    // the original myLambda will be restored at the end of this function so that thisStage will be changed
    // by the function checkForTransformation()
    if (myLambda > thisStage) {
      myLambda = thisStage;
      myTopo->molecules[T].lambda = thisStage - 0.000001;}
    // if myLambda is less than oldStage then temporarily set myLambda to be just more than oldStage;
    // the original myLambda will be restored at the end of this function so that thisStage will be changed
    // by the function checkForTransformation()
    else if (myLambda < oldStage) {
      myLambda = oldStage;
      myTopo->molecules[T].lambda = oldStage + 0.000001;}
    // if myLambda is in between thisStage and oldStage then we don't need to adjust myLambda  
    else myTopo->molecules[T].lambda = myLambda;

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the atomic charges and masses
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // the new mass of the molecule
    Real newMolMass = 0.0;

    // loop over all atoms on molecule T
    for (vector<int>::iterator iter = myTopo->molecules[T].atoms.begin();
         iter != myTopo->molecules[T].atoms.end(); iter++) {

      // this is the correct stage for this atom
      // the index # into the atomTypes array
      int I = myTopo->atoms[*iter].type;

      // the atom type name of this atom
      string myName(myTopo->atomTypes[I].name);
      
      // find the proper iSGAtomType element
      bool Found = false;
      unsigned int myISGAtomType = 0;
      while (!Found) {
        if (myName == myAtomTypes[myISGAtomType].name) Found = true;
        else myISGAtomType++;
      }

      // The mass only needs to be updated if we are at the proper transformation stage for this atom
      if (myTopo->atoms[*iter].stageNumber == thisStage)        
        myTopo->atoms[*iter].scaledMass = myAtomTypes[myISGAtomType].mass[NewType] * (myLambda - oldStage)
          + myAtomTypes[myISGAtomType].mass[OldType] * (thisStage - myLambda);

      // update the atom's charge and deltaQ
      int o = OldType;
      int n = NewType;
      int s = oldStage;
      myTopo->atoms[*iter].scaledCharge = myTransformMaps[myISGAtomType].new_charge[o][n][s] * (myLambda - oldStage)
        + myTransformMaps[myISGAtomType].old_charge[o][n][s] * (thisStage - myLambda);
      myTopo->atoms[*iter].scaledCharge *= Constant::SQRTCOULOMBCONSTANT;

      // update the molecular mass
      newMolMass += myTopo->atoms[*iter].scaledMass;
      
#ifdef DEBUG_MODIFIER_ATOM
        report.precision(8);
        report << hint << "____________________________________________________________________" << endr;
        report << hint << "Looking at atom # " << (*iter) << ", type = "
               << myTopo->atomTypes[myTopo->atoms[*iter].type].name << endr;
        report << hint << "Atom's stage number: " << myTopo->atoms[*iter].stageNumber << ", thisStage = "
               << thisStage << ", Lambda = " << myLambda << endr;
        report << hint << "New m and q = " << myTopo->atoms[*iter].scaledMass << ", " << myTopo->atoms[*iter].scaledCharge / Constant::SQRTCOULOMBCONSTANT << endr;
        report << hint << "DeltaM = " << myTopo->atoms[*iter].deltaM << ", DeltaQ = " << myTopo->atoms[*iter].deltaQ / Constant::SQRTCOULOMBCONSTANT << endr;
        report << hint << "____________________________________________________________________" << endr;
#endif
    } // end loop over atoms

    // update the mass of molecule T
    myTopo->molecules[T].mass = newMolMass;
        
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the BOND parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all bonds on molecule T
    for (vector<int>::iterator iter = myTopo->molecules[T].bondList.begin();
	 iter != myTopo->molecules[T].bondList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = myTopo->atoms[myTopo->bonds[*iter].atom1].stageNumber;
      int atom2_stage = myTopo->atoms[myTopo->bonds[*iter].atom2].stageNumber;

      // determine which stage this bond should be transformed in
      // (it should be the LARGER of the two stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage);

      // update these parameters only if myStage = thisStage
      if (myStage == thisStage) {      
        // find the index # of the proper iSGBond element
        int myISGBond = myTopo->bonds[*iter].iSGmodifierIndex;
      
        // update the force constant, k, and rest distance, r0
        myTopo->bonds[*iter].springConstant = myBonds[myISGBond].forceConstant[NewType] * (myLambda - oldStage)
          + myBonds[myISGBond].forceConstant[OldType] * (thisStage - myLambda);
        myTopo->bonds[*iter].restLength = myBonds[myISGBond].distance[NewType] * (myLambda - oldStage)
          + myBonds[myISGBond].distance[OldType] * (thisStage - myLambda);

        // if we are using rigid bonds then we must loop over the constraint list to find
        // the constraint corresponding to this bond
        if (myTopo->bondRattleShakeConstraints.size() != 0) {

          // get the atom ID#s of the atoms in this bond
          int myAtom1 = myTopo->bonds[*iter].atom1;
          int myAtom2 = myTopo->bonds[*iter].atom2;

          // loop over the constraint list
          for (unsigned int shakeIter = 0; shakeIter < myTopo->bondRattleShakeConstraints.size(); shakeIter++) {

            // get the atom ID#s of the atoms in this constraint
            int shakeAtom1 = myTopo->bondRattleShakeConstraints[shakeIter].atom1;
            int shakeAtom2 = myTopo->bondRattleShakeConstraints[shakeIter].atom2;
            
            // if this is the correct constraint the update the restLength
            if (myAtom1 == shakeAtom1 && myAtom2 == shakeAtom2)
              myTopo->bondRattleShakeConstraints[shakeIter].restLength = myTopo->bonds[*iter].restLength;
          } // end loop over constraints
        } // end if statement

#ifdef DEBUG_MODIFIER_BOND
        report.precision(8);
        report << hint << "Looking at bond #" << (*iter) << endr;
          report << hint << "New k and r0 = " << myTopo->bonds[*iter].springConstant << ", " << myTopo->bonds[*iter].restLength << endr;
#endif
      } // end if statement
    } // end loop over bonds

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the ANGLE parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all angles on molecule T
    for (vector<int>::iterator iter = myTopo->molecules[T].angleList.begin();
	 iter != myTopo->molecules[T].angleList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = myTopo->atoms[myTopo->angles[*iter].atom1].stageNumber;
      int atom2_stage = myTopo->atoms[myTopo->angles[*iter].atom2].stageNumber;
      int atom3_stage = myTopo->atoms[myTopo->angles[*iter].atom3].stageNumber;
        
      // determine which stage this bond should be transformed in
      // (it should be the LARGEST of the stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage,atom3_stage);
       
      // update these parameters only if myStage = thisStage
      if (myStage == thisStage) {          
        // find the index # of the proper iSGAngle element
        int myISGAngle = myTopo->angles[*iter].iSGmodifierIndex;

        // update the force constant, k, and the rest angle, theta0
        myTopo->angles[*iter].forceConstant = myAngles[myISGAngle].forceConstant[NewType] * (myLambda - oldStage)
          + myAngles[myISGAngle].forceConstant[OldType] * (thisStage - myLambda);
        myTopo->angles[*iter].ureyBradleyConstant = myAngles[myISGAngle].k_ub[NewType] * (myLambda - oldStage)
            + myAngles[myISGAngle].k_ub[OldType] * (thisStage - myLambda);
        myTopo->angles[*iter].restAngle = myAngles[myISGAngle].angleval[NewType] * (myLambda - oldStage)
          + myAngles[myISGAngle].angleval[OldType] * (thisStage - myLambda);
        myTopo->angles[*iter].ureyBradleyRestLength = myAngles[myISGAngle].r_ub[NewType] * (myLambda - oldStage)
          + myAngles[myISGAngle].r_ub[OldType] * (thisStage - myLambda);

#ifdef DEBUG_MODIFIER_ANGLE
        report.precision(8);
        report << hint << "Looking at angle # " << (*iter) << endr;
        report << hint << "Index # " << myISGAngle << endr;
        report << hint << "New k and theta0 = " << myTopo->angles[*iter].forceConstant << ", " << myTopo->angles[*iter].restAngle << endr;
        report << hint << "New k and theta0 for UB = " << myTopo->angles[*iter].ureyBradleyConstant 
               << ", " << myTopo->angles[*iter].ureyBradleyRestLength << endr;
#endif
      } // end if statement
    } // end loop over angles
    
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the DIHEDRAL parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all dihedrals on molecule T
    for (vector<int>::iterator iter = myTopo->molecules[T].dihedralList.begin();
	 iter != myTopo->molecules[T].dihedralList.end(); iter++) {
      
      // get the transformation stage #'s for these atoms
      int atom1_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom1].stageNumber;
      int atom2_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom2].stageNumber;
      int atom3_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom3].stageNumber;
      int atom4_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom4].stageNumber;
        
      // determine which stage this bond should be transformed in
      // (it should be the LARGEST of the stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);
       
      // update these parameters only if myStage = thisStage
      if (myStage == thisStage) {      
        // find the index # of the proper iSGDihedral element
        int myISGDihedral = myTopo->dihedrals[*iter].iSGmodifierIndex;
        
        for (int i=0; i<myTopo->dihedrals[*iter].multiplicity; i++) {
          // update the force constants
          myTopo->dihedrals[*iter].forceConstant[i] =
            myDihedrals[myISGDihedral].forceConstant[NewType][i] * (myLambda - oldStage)
            + myDihedrals[myISGDihedral].forceConstant[OldType][i] * (thisStage - myLambda);
 
          // update the phase shift angles
          myTopo->dihedrals[*iter].phaseShift[i] =
            myDihedrals[myISGDihedral].phaseShift[NewType][i] * (myLambda - oldStage)
            + myDihedrals[myISGDihedral].phaseShift[OldType][i] * (thisStage - myLambda);

#ifdef DEBUG_MODIFIER_TORSION
          report.precision(8);
          report << hint << "Looking at dihedral # " << (*iter) << endr;
          report << hint << "New k and P.S. = " << myTopo->dihedrals[*iter].forceConstant[i] << ", " << myTopo->dihedrals[*iter].phaseShift[i] << endr;
          report << hint << "DeltaK = " << myTopo->dihedrals[*iter].DeltaK[i] << endr;
          report << hint << "DeltaPhase = " << myTopo->dihedrals[*iter].DeltaPhase[i] << endr;
#endif
        } // end loop over i
      } // end if statement
    } // end loop over dihedrals

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // update the IMPROPER parameters
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // loop over all impropers on molecule T
    for (vector<int>::iterator iter = myTopo->molecules[T].improperList.begin();
	 iter != myTopo->molecules[T].improperList.end(); iter++) {

      // get the transformation stage #'s for these atoms
      int atom1_stage = myTopo->atoms[myTopo->impropers[*iter].atom1].stageNumber;
      int atom2_stage = myTopo->atoms[myTopo->impropers[*iter].atom2].stageNumber;
      int atom3_stage = myTopo->atoms[myTopo->impropers[*iter].atom3].stageNumber;
      int atom4_stage = myTopo->atoms[myTopo->impropers[*iter].atom4].stageNumber;
        
      // determine which stage this bond should be transformed in
      // (it should be the LARGEST of the stage #s, if they are different)
      int myStage = max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);

      // update these parameters only if myStage = thisStage
      if (myStage == thisStage) {      
        // find the index # of the proper iSGImproper element
        int myISGImproper = myTopo->impropers[*iter].iSGmodifierIndex;
      
        // update the force constant 
        myTopo->impropers[*iter].forceConstant[0] =
          myImpropers[myISGImproper].forceConstant[NewType] * (myLambda - oldStage)
          + myImpropers[myISGImproper].forceConstant[OldType] * (thisStage - myLambda);

        // update the phase shift angle
        myTopo->impropers[*iter].phaseShift[0] =
          myImpropers[myISGImproper].phaseShift[NewType] * (myLambda - oldStage)
          + myImpropers[myISGImproper].phaseShift[OldType] * (thisStage - myLambda);

#ifdef DEBUG_MODIFIER_TORSION
        report.precision(8);
        report << hint << "Looking at improper # " << (*iter) << endr;
        report << hint << "New k and P.S. = " << myTopo->impropers[*iter].forceConstant[0] << ", " << myTopo->impropers[*iter].phaseShift[0] << endr;
#endif
      } // end if statement
    } // end loop over impropers

    // restore the original myLambda
    myLambda = PreModLambda;
  } // end function modifyForces()

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  checkForTransformation().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::checkForTransformation(const vector<iSGPAR::Bond> &myBonds,
                                             const vector<iSGPAR::Angle> &myAngles,
                                             const vector<iSGPAR::Dihedral> &myDihedrals,
                                             const vector<iSGPAR::Improper> &myImpropers,
                                             const vector<iSGPAR::AtomType> &myAtomTypes,
                                             const vector<ModifierISG::TransformMap> &myTransformMaps) {

    Transformed = false;

    // determine which transformation state the molecule is currently in
    int newStage = static_cast<int>(floor(myLambda)) + 1;

    // determine if the molecule has been completely transformed 
    if (myLambda > (myNumStages - 0.002) ) {
      //  add to the # of molecules of NewType and subtract from the 
      //  # of molecules of OldType
      N[NewType]++;
      myTopo->iSGNumMols[NewType]++;
      N[OldType]--;
      myTopo->iSGNumMols[OldType]--;
      
      // set Lambda for this molecule to be equal to 0 and update
      // the molecule's type
      myTopo->molecules[T].lambda = 0.0;
      myTopo->molecules[T].type = NewType;
      myTopo->molecules[T].newtype = NewType;
      FinalType = NewType;
      Transformed = true;
    }
    else if (myLambda < 0.002) {
      // set Lambda for this molecule to be equal to 0 and update
      // the molecule's type
      myTopo->molecules[T].lambda = 0.0;
      myTopo->molecules[T].type = OldType;
      myTopo->molecules[T].newtype = OldType;
      FinalType = OldType;
      Transformed = true;
    }   
    else if (newStage != thisStage) {

      // the transformation attempt has not yet completed,
      // but it has reached a new stage so we need to update
      // the force constants which are used for DeltaMu calculations

      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the stomic charges and masses
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      mylnMassRatio = 0.0;

      // loop over all atoms on molecule T
      for (std::vector<int>::iterator iter = myTopo->molecules[T].atoms.begin();
           iter != myTopo->molecules[T].atoms.end(); iter++) {
      
        // the index # into the atomTypes array
        int I = myTopo->atoms[*iter].type;
        
        // the atom type name of this atom
        string myName(myTopo->atomTypes[I].name);
        
        // find the proper iSGAtomType element
        bool Found = false;
        unsigned int myISGAtomType = 0;
        while (!Found) {
          if (myName == myAtomTypes[myISGAtomType].name) Found = true;
          else myISGAtomType++;
        }
        
        // update deltaM, deltaQ, Qold, and Qnew
        if (myTopo->atoms[*iter].stageNumber == thisStage) myTopo->atoms[*iter].deltaM = 0.0;

        // to make sure we compute the intramolecular Coulomb force correctly, we must
        // assign the proper charge to Qold and Qnew
        myTopo->atoms[*iter].Qold = myTransformMaps[myISGAtomType].new_charge[OldType][NewType][newStage-1];
        myTopo->atoms[*iter].Qnew = myTransformMaps[myISGAtomType].old_charge[OldType][NewType][newStage-1];
        myTopo->atoms[*iter].deltaQ = myTransformMaps[myISGAtomType].new_charge[OldType][NewType][newStage-1]
          - myTransformMaps[myISGAtomType].old_charge[OldType][NewType][newStage-1];
        myTopo->atoms[*iter].Qold *= Constant::SQRTCOULOMBCONSTANT;
        myTopo->atoms[*iter].Qnew *= Constant::SQRTCOULOMBCONSTANT;          
        myTopo->atoms[*iter].deltaQ *= Constant::SQRTCOULOMBCONSTANT;

        if (myTopo->atoms[*iter].stageNumber == newStage) {
          myTopo->atoms[*iter].deltaM = myAtomTypes[myISGAtomType].mass[NewType]
            - myAtomTypes[myISGAtomType].mass[OldType];

          mylnMassRatio += (3 * kbT * 0.5) * (log(myAtomTypes[myISGAtomType].mass[OldType])
            - log(myAtomTypes[myISGAtomType].mass[NewType]));
        } // end if-else statements
      } // end loop over atoms
            
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the BOND parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all bonds on molecule T
      for (std::vector<int>::iterator iter = myTopo->molecules[T].bondList.begin();
           iter != myTopo->molecules[T].bondList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = myTopo->atoms[myTopo->bonds[*iter].atom1].stageNumber;
        int atom2_stage = myTopo->atoms[myTopo->bonds[*iter].atom2].stageNumber;

        // determine which stage this bond should be transformed in
        // (it should be the LARGER of the two stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage);
                
        // find the index # of the proper iSGBond element
        int myISGBond = myTopo->bonds[*iter].iSGmodifierIndex;
        
        // update DeltaK and DeltaR0
        if (myStage == thisStage) { 
          // this stage is complete, so we set the parameters to zero
          myTopo->bonds[*iter].DeltaK = 0.0;
          myTopo->bonds[*iter].DeltaR0 = 0.0;
        }
        else if (myStage == newStage) {
          // we are now starting this new stage, so these parameters must be updated
          myTopo->bonds[*iter].DeltaK = myBonds[myISGBond].forceConstant[NewType]
            - myBonds[myISGBond].forceConstant[OldType];
          myTopo->bonds[*iter].DeltaR0 = myBonds[myISGBond].distance[NewType]
            - myBonds[myISGBond].distance[OldType];
        } // end if-else statements
      } // end loop over bonds
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the ANGLE parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all angles on molecule T
      for (std::vector<int>::iterator iter = myTopo->molecules[T].angleList.begin();
           iter != myTopo->molecules[T].angleList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = myTopo->atoms[myTopo->angles[*iter].atom1].stageNumber;
        int atom2_stage = myTopo->atoms[myTopo->angles[*iter].atom2].stageNumber;
        int atom3_stage = myTopo->atoms[myTopo->angles[*iter].atom3].stageNumber;
        
        // determine which stage this bond should be transformed in
        // (it should be the LARGEST of the stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage,atom3_stage);
                           
        // find the index # of the proper iSGAngle element
        int myISGAngle = myTopo->angles[*iter].iSGmodifierIndex;
        
        // update DeltaK and DeltaTheta0
        if (myStage == thisStage) {
          // this stage is complete, so we set the parameters to zero
          myTopo->angles[*iter].DeltaK = 0.0;
          myTopo->angles[*iter].DeltaTheta0 = 0.0;
          myTopo->angles[*iter].Delta_ubK = 0.0;
          myTopo->angles[*iter].Delta_ubR0 = 0.0;
        }
        else if (myStage == newStage) {
          // we are now starting this new stage, so these parameters must be updated
          myTopo->angles[*iter].DeltaK = myAngles[myISGAngle].forceConstant[NewType]
            - myAngles[myISGAngle].forceConstant[OldType];
          myTopo->angles[*iter].DeltaTheta0 = myAngles[myISGAngle].angleval[NewType]
            - myAngles[myISGAngle].angleval[OldType];
          myTopo->angles[*iter].Delta_ubK = myAngles[myISGAngle].k_ub[NewType]
            - myAngles[myISGAngle].k_ub[OldType];
          myTopo->angles[*iter].Delta_ubR0 = myAngles[myISGAngle].r_ub[NewType]
            - myAngles[myISGAngle].r_ub[OldType];
        } // end if-else statements
      } // end loop over angles
      
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the DIHEDRAL parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all dihedrals on molecule T
      for (std::vector<int>::iterator iter = myTopo->molecules[T].dihedralList.begin();
           iter != myTopo->molecules[T].dihedralList.end(); iter++) {
        
        // get the transformation stage #'s for these atoms
        int atom1_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom1].stageNumber;
        int atom2_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom2].stageNumber;
        int atom3_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom3].stageNumber;
        int atom4_stage = myTopo->atoms[myTopo->dihedrals[*iter].atom4].stageNumber;
        
        // determine which stage this bond should be transformed in
        // (it should be the LARGEST of the stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);
                
        // find the index # of the proper iSGDihedral element
        int myISGDihedral = myTopo->dihedrals[*iter].iSGmodifierIndex;
        
        for (int i=0; i<myTopo->dihedrals[*iter].multiplicity; i++) {

          // update DeltaK and DeltaPhase
          if (myStage == thisStage) {
            // this stage is complete, so we set the parameters to zero
            myTopo->dihedrals[*iter].DeltaK[i] = 0.0;
            myTopo->dihedrals[*iter].DeltaPhase[i] = 0.0;
          }
          else if (myStage == newStage) {
            // we are now starting this new stage, so these parameters must be updated
            myTopo->dihedrals[*iter].DeltaK[i] = myDihedrals[myISGDihedral].forceConstant[NewType][i]
              - myDihedrals[myISGDihedral].forceConstant[OldType][i];
            myTopo->dihedrals[*iter].DeltaPhase[i] = myDihedrals[myISGDihedral].phaseShift[NewType][i]
              - myDihedrals[myISGDihedral].phaseShift[OldType][i];
          } // end if-else statements          
        } // end loop over i  
      } // end loop over dihedrals
    
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      // update the IMPROPER parameters
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     
      // loop over all impropers on molecule T
      for (std::vector<int>::iterator iter = myTopo->molecules[T].improperList.begin();
           iter != myTopo->molecules[T].improperList.end(); iter++) {

        // get the transformation stage #'s for these atoms
        int atom1_stage = myTopo->atoms[myTopo->impropers[*iter].atom1].stageNumber;
        int atom2_stage = myTopo->atoms[myTopo->impropers[*iter].atom2].stageNumber;
        int atom3_stage = myTopo->atoms[myTopo->impropers[*iter].atom3].stageNumber;
        int atom4_stage = myTopo->atoms[myTopo->impropers[*iter].atom4].stageNumber;
        
        // determine which stage this bond should be transformed in
        // (it should be the LARGEST of the stage #s, if they are different)
        int myStage =  max(atom1_stage,atom2_stage,atom3_stage,atom4_stage);
                           
        // find the index # of the proper iSGImproper element
        int myISGImproper = myTopo->impropers[*iter].iSGmodifierIndex;
        

        // update DeltaK and DeltaPhase
        if (myStage == thisStage) {
          // this stage is complete, so we set the parameters to zero        
          myTopo->impropers[*iter].DeltaK[0] = 0.0;
          myTopo->impropers[*iter].DeltaPhase[0] = 0.0;
        }
        else if (myStage == newStage) {
          // we are now starting this new stage, so these parameters must be updated
          myTopo->impropers[*iter].DeltaK[0] = myImpropers[myISGImproper].forceConstant[NewType]
            - myImpropers[myISGImproper].forceConstant[OldType];
          myTopo->impropers[*iter].DeltaPhase[0] = myImpropers[myISGImproper].phaseShift[NewType]
            - myImpropers[myISGImproper].phaseShift[OldType];
        } // end if-else statements
      } // end loop over impropers
      
      // Lastly, update the transformation stage indicator and the ideal gas DeltaMu
      thisStage = newStage;
      myDMuIG = myDeltaMuIG[OldType][NewType][thisStage-1];
    }

    // if the transformation attempt has completed then compute the
    // average deviation in the conserved quantity
    if (Transformed) {     
      // compute the average of the conserved quantity
      AveCQ /= static_cast<Real>(NumTransSteps);
      AveCQSq /= static_cast<Real>(NumTransSteps);

      // compute the fluctuation in the CQ
      (*myEnergies)[ScalarStructure::CQFLUCTUATION] = AveCQSq - power(AveCQ,2);
      (*myEnergies)[ScalarStructure::DELTATIME] = NumTransSteps;
      
      // output the system properties to the AllEnergiesFile
      myEnergies->output(true);

      // reset the transformation stage indicator to stage 1
      thisStage = 1;
    } // end if (Transformed) statement
  } // end function
 
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // run
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::run(int numTimesteps) {
    for (int i = 0; i < numTimesteps; i++) {

      // integrate the equations of motion
      preStepModify();
      doHalfKick();
      doDriftOrNextIntegrator();
      calculateForces();
      do2ndHalfKick();
      postStepModify();

      //  add to the conserved quantity averages
      Real MyCQ = kineticEnergy(myTopo, myVelocities) + myEnergies->potentialEnergy()
        + (*myEnergies)[ScalarStructure::INTEGRATOR];
      AveCQ += MyCQ;
      AveCQSq += power(MyCQ, 2);
    }
  }
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  initialize().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::initialize(GenericTopology *topo,
                                 Vector3DBlock *positions,
                                 Vector3DBlock *velocities, 
                                 ScalarStructure *energies) {

    STSIntegrator::initialize(topo, positions, velocities, energies);

    //  store the total number of atoms and molecules
    NumAtoms = myTopo->atoms.size();
    NumMols = myTopo->molecules.size();
    myNumFree = myTopo->degreesOfFreedom;

    //  determine the # of molecules of each component in the mixture
    N.resize(myNumComp);
    myTopo->iSGNumMols.resize(myNumComp);
    for (unsigned int C=0; C < myNumComp; C++) {
      N[C] = 0;
      myTopo->iSGNumMols[C] = 0;
    }
    for (unsigned int M=0; M < NumMols; M++) {
      N[myTopo->molecules[M].type]++;
      myTopo->iSGNumMols[myTopo->molecules[M].type]++;
    }

    //  Initialize integrator variables to something sane.
    if (T == -1) {
      myEpsilonVel = 0.;
      myEta        = 0.;
      myEtaV       = 0.;
      myEtaVel     = 0.;
      myEtaVolVel  = 0.;
      myLambdaVel  = 0.;

      //  Initialize myVolume
      myVolume = myTopo->getVolume(*positions);
    }
    else {
      // Initialize the volume
      Real fac = myVolume / myTopo->getVolume();
      myTopo->rescaleVolume(fac);
      uncache();

      // Initialize the transforming molecule identities
      myTopo->molecules[T].lambda = myLambda;
      myTopo->molecules[T].type = OldType;
      myTopo->molecules[T].newtype = NewType;

      // Compute the residual and ideal gas chemical potential differences:
      // ideal gas chemical potential difference (using the precomputed chemical potentials)
      myDMuIG = myDeltaMuIG[OldType][NewType][thisStage-1];

      // residual chemical potential difference
      // DeltaMu = kT*ln(f1/f0 * N0/(N1+1))
      myTargetDeltaMu = (myTargetMu[NewType] - myTargetMu[OldType])
	+ kbT * log (static_cast<Real>(N[OldType]) / static_cast<Real>(N[NewType]+1));
      myTargetDeltaMu /= static_cast<Real>(myNumStages);
    }

    //  Compute the fixed barostat mass and thermostat masses.
    Qo = myNumFree * kbT * (myTauT * myTauT);
    Qv = kbT * (myTauV * myTauV);
    W = (myNumFree + 3) * kbT * (myTauP * myTauP);
    Qd = kbT * (myTauD * myTauD);

    initializeForces();
    myEnergies->output(false);
  }
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  NPT RATTLE modifier
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
  Modifier* iSGIntegrator::createRattleModifier(Real eps, int maxIter){
    return (new ModifierNPTRattle<iSGIntegrator>(eps, maxIter, this));
  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  NPT SHAKE modifier
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Modifier* iSGIntegrator::createShakeModifier(Real eps, int maxIter){
    return (new ModifierNPTShake<iSGIntegrator>(eps, maxIter, this));
  }
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  Add the modifiers
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::addModifierAfterInitialize() {
    adoptPreStepModifier(new ModifierPreForceThermostat<iSGIntegrator>(this,1));
    adoptPreStepModifier(new ModifierPreForceBarostat<iSGIntegrator>(this,2));
    adoptPreStepModifier(new iSGModifierPreForceChemostat(this,3));
    adoptPostStepModifier(new iSGModifierPostForceChemostat(this,1));
    adoptPostStepModifier(new ModifierPostForceBarostat<iSGIntegrator>(this,2));
    adoptPostStepModifier(new ModifierPostForceThermostat<iSGIntegrator>(this,3));
    STSIntegrator::addModifierAfterInitialize();
  }
  
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  //  addModifierBeforeInitialize().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::addModifierBeforeInitialize(){
    //    adoptPreForceModifier(new ModifierISGforces(this));
  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // getParameters().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  void iSGIntegrator::getParameters(vector< Parameter >& parameters) const {
    STSIntegrator::getParameters(parameters);
    parameters.push_back(Parameter("components", Value(myNumComp)));
    parameters.push_back(Parameter("temperature", Value(myTargetTemp)));
    parameters.push_back(Parameter("pressure", Value(myTargetPres)));
    parameters.push_back(Parameter("fugacityfrac", Value(myFugacityFrac)));
    parameters.push_back(Parameter("tauT", Value(myTauT)));
    parameters.push_back(Parameter("tauV", Value(myTauV)));
    parameters.push_back(Parameter("tauP", Value(myTauP)));
    parameters.push_back(Parameter("tauD", Value(myTauD)));
  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // doMake().
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  STSIntegrator* iSGIntegrator::doMake(string&, const vector<Value>& values,ForceGroup* fg)const{
    return new iSGIntegrator(values[0], values[1], values[2], values[3], values[4],
			     values[5], values[6], values[7], values[8], fg );
  }

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // finalXSCs
  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  XSC iSGIntegrator::getXSC() const {

    // create an XSC object to store the final xsc values
    XSC xsc;

    // set the integrator type
    xsc.simType = "iSGVerlet";
      
    // for chemostat
    xsc.Lambda = myLambda;
    xsc.Lambda_vel = myLambdaVel;
    xsc.myMolecule = T+1;
    xsc.old_type = OldType;
    xsc.new_type = NewType;

    // for thermostat
    xsc.Eta = myEta;
    xsc.EtaVol = myEtaV;
    xsc.Eta_vel = myEtaVel;
    xsc.EtaVol_vel = myEtaVolVel;
      
    // for barostat
    xsc.Vol = myVolume; 
    xsc.Epsilon_vel = myEpsilonVel;
    return xsc;
  }
}
