/* -*- c++ -*- */
#ifndef COULOMBMULTIGRIDDIRECTTABLEFORCE_H
#define COULOMBMULTIGRIDDIRECTTABLEFORCE_H

#include "GenericTopology.h"
#include "ScalarStructure.h"
#include "ExclusionTable.h"
#include "Parameter.h"
#include "mathutilities.h"
#include <string>
#include "CoulombMultiGridDirectTableForceBase.h"

namespace ProtoMol {

  //_________________________________________________________________ CoulombMultiGridDirectTableForce
  template<class TKernel>  
  class CoulombMultiGridDirectTableForce : private CoulombMultiGridDirectTableForceBase {
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Replacement potential of the real part for MultiGrid or PME.
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
    enum {DEFAULT__SIZE=20000};
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Constructors, destructors, assignment
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  public:
    CoulombMultiGridDirectTableForce():myTable(NULL),myS(0.0),myN(0){}
    CoulombMultiGridDirectTableForce(Real s, unsigned int n):myTable(new Real[2*n]),myS(s),myRS(1.0/s),myFac(((Real)n-2.0)/s),myN(n){
      for(unsigned int i=0;i<myN;++i){
	Real r = (i+(i==0?0.01:0.0))/myFac;      
	myTable[i*2  ] = ( TKernel::kernel(r)  - TKernel::smooth(r,myS,myRS)); 
	myTable[i*2+1] = (-TKernel::dKernel(r) + TKernel::dSmooth(r,myS,myRS))*1.0/r;	
      }      
      Report::report <<Report::hint<<"Using MultiGrid direct approximation with "<<n<<" points, range ["<<1.0/myFac<<","<<(Real)n/myFac<<"[, [0,1,"<<myN-1<<"] = ("<<myTable[0]<<","<<myTable[1]<<"),("<<myTable[2]<<","<<myTable[3]<<"),("<<myTable[(int)myN*2-2]<<","<<myTable[(int)myN*2-1]<<")."<<Report::endr;
    }
    CoulombMultiGridDirectTableForce(CoulombMultiGridDirectTableForce const& other):myTable(new Real[2*other.myN]),
										myS(other.myS),
										myRS(other.myRS),
										myFac(other.myFac),
										myN(other.myN){
      for(unsigned int i=0;i<myN*2;++i)
	myTable[i] = other.myTable[i];          
    }
    CoulombMultiGridDirectTableForce& operator=(CoulombMultiGridDirectTableForce const& other){
      if(&other != this){
	if(myTable != NULL)
	  delete [] myTable;
	myS            = other.myS;
	myRS           = other.myRS;
	myFac          = other.myFac;
	myN            = other.myN;
	myTable        = new Real[2*myN];
	for(unsigned int i=0;i<myN*2;++i)
	  myTable[i] = other.myTable[i];    
      }
      return *this;
    }
    ~CoulombMultiGridDirectTableForce(){if(myTable != NULL)  delete [] myTable; myTable = NULL;}
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // New methods of class CoulombMultiGridDirectTableForce
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  public:
    void operator()(Real &energy, 
		    Real &force, 
		    Real distSquared,
		    Real /*rDistSquared*/,
		    const Vector3D &, 
		    const GenericTopology* topo, 
		    int atom1, int atom2,
		    ExclusionClass excl) const {

      Real qq = topo->atoms[atom1].scaledCharge * topo->atoms[atom2].scaledCharge;
      if (topo->coulombScalingFactor != 1.0 && excl == EXCLUSION_MODIFIED) 
	qq *= topo->coulombScalingFactor;
      Real n = sqrt(distSquared)*myFac;
      int index = (int)n;
      Real b = n-index;
      //Report::report << index <<","<<sqrt(distSquared);
      index <<=1;
      //Report::report <<","<< index <<","<<myN<<","<<(unsigned long long)my<< Report::endr;
      Real a = 1.0-b;
      energy = (myTable[index  ]*a+myTable[index+2]*b)*qq;
      force  = (myTable[index+1]*a+myTable[index+3]*b)*qq;
    }

    static void accumulateEnergy(ScalarStructure* energies, Real energy) {
      (*energies)[ScalarStructure::COULOMB] += energy;
    }

    static Real getEnergy(const ScalarStructure* energies) {return  (*energies)[ScalarStructure::COULOMB];}

    // Parsing
    static std::string getId() {return keyword+" -kernel "+TKernel::keyword;}
    static unsigned int getParameterSize() {return 2;}
    void getParameters(std::vector<Parameter>&parameters) const{
      parameters.push_back(Parameter("-s",Value(myS,ConstraintValueType::Positive()),Text("MG smoothing distance")));
      parameters.push_back(Parameter("-Size",Value(myN,ConstraintValueType::Positive()),static_cast<unsigned int>(DEFAULT__SIZE),Text("size of look up")));
    }
    static CoulombMultiGridDirectTableForce make(std::string&, const std::vector<Value>& values){return CoulombMultiGridDirectTableForce(values[0],values[1]);}
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // My data members
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  private:
    Real* myTable;
    Real myS;
    Real myRS;
    Real myFac;
    unsigned int myN;
  };

  //______________________________________________________________________ INLINES

}
#endif /* COULOMBMULTIGRIDDIRECTTABLEFORCE_H */
