#include "OutputEnergies.h"
#include "Configuration.h"
#include "GenericTopology.h"
#include "ScalarStructure.h"
#include "topologyutilities.h"
#include "OutputCache.h"
#include "inputValueDefinitions.h"

#include <iomanip>

using namespace ProtoMol::Report;

using std::string;
using std::vector;
using std::setw;
using std::endl;
using std::flush;
using std::setprecision;
using std::setiosflags;
using std::resetiosflags;
using std::ofstream;

namespace ProtoMol {
  //________________________________________________________ Output
  const string  OutputEnergies::keyword("allEnergiesFile");

  OutputEnergies::OutputEnergies():OutputFile(),myDoMolecularTemperature(false),myDoShadow(false){}
  OutputEnergies::OutputEnergies(const string& filename, int freq, bool doMolTemp, bool doShadow):OutputFile(filename,freq),myDoMolecularTemperature(doMolTemp),myDoShadow(doShadow){}

  void OutputEnergies::doInitialize(){
    ofstream allEnergiesHeaderFile(string(myFilename + ".header").c_str(), std::ios::out | std::ios::trunc);
    if(!allEnergiesHeaderFile)
      report << error <<" Can not open \'"<<myFilename<<".header\' for "<<getId()<<"."<<endr;

    allEnergiesHeaderFile << setw(18)
			  << "Time(fs)" << " "
			  << setw(14)
			  << "E_potential" << " "
			  << setw(14)
			  << "E_kinetic" << " "
			  << setw(14)
			  << "E_total" << " "
			  << setw(14)
			  << "Temperature" << " "
			  << setw(14)
			  << "E_bond" << " "
			  << setw(14)
			  << "E_angle" << " "
			  << setw(14)
			  << "E_dihedral" << " "
			  << setw(14)
			  << "E_improper" << " "
			  << setw(14)
			  << "E_VdW" << " "
			  << setw(14)
			  << "E_coulomb" << " "
			  << setw(14)
			  << "E_other" << " ";
    if(myEnergies->virial())
      allEnergiesHeaderFile << setw(14) 
			    << "Pressure(bar)" << " ";
    if(myEnergies->molecularVirial())
      allEnergiesHeaderFile << setw(14)
			    << "Mol_Pres(bar)" << " ";
    if(myDoMolecularTemperature)
      allEnergiesHeaderFile << setw(14)
			    << "Mol_Temp(K)" << " ";
    allEnergiesHeaderFile << setw(14)
			  << "Volume(A^3)" << " ";
    if(myDoShadow)
      allEnergiesHeaderFile << setw(14)
			    << "E_shadow";
    
    allEnergiesHeaderFile << endl;
    allEnergiesHeaderFile.close();
    open();
    close();

  }

  void OutputEnergies::doRun(int){
    if(!reopen())
      report << error <<" Can not open \'"<<myFilename<<"\' for "<<getId()<<"."<<endr;

    myFile  << resetiosflags(std::ios::showpoint |  std::ios::fixed | std::ios::floatfield)
	    << setw(18)
	    << setprecision(2)
	    << setiosflags(std::ios::showpoint | std::ios::fixed)
	    << myCache->time() << " "
	    << resetiosflags(std::ios::showpoint |  std::ios::fixed | std::ios::floatfield)
	    << setiosflags(std::ios::floatfield)
	    << setprecision(8)
	    << setw(14)
	    << myCache->potentialEnergy() << " "
	    << setw(14)
	    << myCache->kineticEnergy() << " "
	    << setw(14)
	    << myCache->totalEnergy() << " "
	    << setw(14)
	    << myCache->temperature() << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::BOND] << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::ANGLE] << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::DIHEDRAL] << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::IMPROPER] << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::LENNARDJONES] << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::COULOMB] << " "
	    << setw(14)
	    << (*myEnergies)[ScalarStructure::OTHER]<< " ";
    if(myEnergies->virial())
      myFile << setw(14)
	     << myCache->pressure();
    if(myEnergies->molecularVirial())
      myFile << setw(14) 
	     << myCache->molecularPressure();
    if(myDoMolecularTemperature)
      myFile << setw(14) 
	     << myCache->molecularTemperature();
    myFile << setw(14)
	   << myCache->volume();
    if(myDoShadow)
      myFile << setw(14)
	     << (*myEnergies)[ScalarStructure::SHADOW] ;
    myFile<< endl;

    reclose();
  }

  void OutputEnergies::doFinalize(int){
    close();
  }

  Output* OutputEnergies::doMake(string&, const vector<Value>& values) const{
    return (new OutputEnergies(values[0],values[1],values[2],values[3]));
  }

  void OutputEnergies::getParameters(vector<Parameter> &parameter) const{
    parameter.push_back(Parameter(getId(),Value(myFilename,ConstraintValueType::NotEmpty())));
    parameter.push_back(Parameter("energiesOutputFreq",Value(myOutputFreq,ConstraintValueType::Positive())));
    parameter.push_back(Parameter("molecularTemperature",Value(myDoMolecularTemperature),false));
    parameter.push_back(Parameter("shadowEnergy",Value(myDoShadow),false));
  }

  bool OutputEnergies::adjustWithDefaultParameters(vector<Value>& values, const Configuration* config) const{
    if(!checkParameterTypes(values))
      return false;
    if(config->valid(InputOutputfreq::keyword) && !values[1].valid())
      values[1] = (*config)[InputOutputfreq::keyword];
    return checkParameters(values);
  }

}
