#include "MCIntegrator.h"
#include "Report.h"
#include "GenericTopology.h"
#include "Vector3DBlock.h"
#include "ScalarStructure.h"
#include "topologyutilities.h"
#include "ForceGroup.h"

using namespace ProtoMol::Report;
using std::vector;

namespace ProtoMol {
  //____________________________________________________________ MCIntegrator

  MCIntegrator::MCIntegrator()    
    : MTSIntegrator(),
      myRandomCycLen(false),
      myInitialTemperature(-1),
      myOldKineticEnergy(-1),
      myOldPositions(NULL),
      myOldVelocities(NULL),
      myOldEnergies(NULL){
  }

  //  ---------------------------------------------------------------------  //

  MCIntegrator::MCIntegrator(int cycles,
			     bool randomCycLen,
			     Real initialTemperature,
			     ForceGroup *overloadedForces,
			     StandardIntegrator *nextIntegrator)
    
    : MTSIntegrator(cycles,overloadedForces,nextIntegrator),
      myRandomCycLen(randomCycLen),
      myInitialTemperature(initialTemperature),
      myOldKineticEnergy(-1),
      myOldPositions(new Vector3DBlock),
      myOldVelocities(new Vector3DBlock),
      myOldEnergies(new ScalarStructure) {
    if(overloadedForces->anyForces()){
      report << error << "Monte Carlo integrators do not expect any forces!"<<endr;
    }

  }
  
  //  ---------------------------------------------------------------------  //

  MCIntegrator::~MCIntegrator(){
    if(myOldPositions != NULL)
      delete myOldPositions;
    if(myOldVelocities != NULL)
      delete myOldVelocities;
    if(myOldEnergies != NULL)
      delete myOldEnergies;
  }

  //  ---------------------------------------------------------------------  //

  void MCIntegrator::initialize(GenericTopology *topo,
				Vector3DBlock *positions,
				Vector3DBlock *velocities,
				ScalarStructure *energies){

    MTSIntegrator::initialize(topo,positions,velocities,energies);

    //Report::report <<"[MCIntegrator::initialize]"<<Report::endr;

    myOldPositions->resize(positions->size());
    myOldVelocities->resize(velocities->size());


  }

  //  ---------------------------------------------------------------------  //

  void MCIntegrator::run(int numTimesteps) {

    saveValues();

    for(int i = 0; i < numTimesteps; i++) {

      preStepModify();

      perturbSystem();

      //  The metropolis test is based in part on the change in KE from the new
      //  velocities.
      myOldKineticEnergy = kineticEnergy(myTopo,myVelocities);

      if(myRandomCycLen) {
	walk((int)(myCycleLength * (1.3 - 0.6 * randomNumber()))); 
      }
      else{
	walk(myCycleLength); 
      }

      if(metropolisTest()) {
	saveValues();

        report << debug(1) << "Move accepted" << endr;

      }
      else {
	restoreValues();

        report << debug(1) << "Move rejected" << endr;

      } 
      postStepModify();
    }  

  }

  //  ---------------------------------------------------------------------  //

  bool MCIntegrator::metropolisTest() {

    return metropolisTest( myEnergies->potentialEnergy() +
                           kineticEnergy(myTopo,myVelocities),
			   myOldEnergies->potentialEnergy() +
                           myOldKineticEnergy );
  }

  //  ---------------------------------------------------------------------  //

  bool MCIntegrator::metropolisTest( Real newEnergy, Real oldEnergy ) {

    Real deltaE = newEnergy - oldEnergy;

    if(deltaE < 0)  
      return true;

    Real acceptProb = exp( -deltaE /
                           ( Constant::BOLTZMANN * getInitialTemperature() ) );

    report << debug(2) << "probability = " << acceptProb << endr;

    Real r = randomNumber();

    return( r < acceptProb );

  }

  //  ---------------------------------------------------------------------  //

  void MCIntegrator::getParameters(vector< Parameter> &parameters) const {
    MTSIntegrator::getParameters(parameters);
    parameters.push_back(Parameter( "randomCycLen", Value( myRandomCycLen ), false ) );
    parameters.push_back(Parameter( "temperature", Value( myInitialTemperature,
                    ConstraintValueType::NotNegative() ) ) );
  }

  //  ---------------------------------------------------------------------  //

  void MCIntegrator::saveValues() {
    (*myOldPositions)  = (*myPositions);
    (*myOldVelocities) = (*myVelocities);
    (*myOldEnergies)   = (*myEnergies);
  }

  //  ---------------------------------------------------------------------  //

  void MCIntegrator::restoreValues() {
    (*myPositions)  = (*myOldPositions);
    (*myVelocities) = (*myOldVelocities);
    (*myEnergies)   = (*myOldEnergies);
    buildMolecularCenterOfMass(myPositions,myTopo);
    buildMolecularMomentum(myVelocities,myTopo);
  }

  //  ---------------------------------------------------------------------  //

  void MCIntegrator::walk(int steps){
    myNextIntegrator->run(steps);
  }

}

