#include "PDBReader.h"

#include "stringutilities.h"
#include "Report.h"
#include "mathutilities.h"

using std::string;
using std::vector;
using std::endl;
using std::find;
using std::stringstream;
using namespace ProtoMol::Report;

//#define DEBUG_PDB

namespace ProtoMol {
  //_________________________________________________________________PDBReader

  PDBReader::PDBReader():
    Reader(),
    myCoords(NULL),myAtoms(NULL){}

  PDBReader::PDBReader(const std::string& filename):
    Reader(filename),
    myCoords(NULL),myAtoms(NULL){}


  PDBReader::~PDBReader(){
    if(myCoords != NULL)
      delete myCoords;
    if(myAtoms != NULL)
      delete myAtoms;
  }

  bool PDBReader::tryFormat(){
    if(!open())
      return false;
    do {
      string record,str;
      record = getline();
      stringstream ss(record);
      ss >> str;
      if("ATOM" == str){
	close();
	return true;
      }
    } while (!myFile.eof());
    myFile.setstate(std::ios::failbit);
    close();        
    return false;
  }

  bool PDBReader::read() {
    if(myCoords == NULL)
      myCoords = new Vector3DBlock();
    if(myAtoms == NULL)
      myAtoms = new std::vector<PDB::PDBAtom>();
    return read(*myCoords,*myAtoms);
  }

  bool PDBReader::read(PDB& pdb){
    return read(pdb.coords,pdb.atoms);
  }
  
  bool PDBReader::read(Vector3DBlock& coords, std::vector<PDB::PDBAtom>& atoms) {
    if(!tryFormat())
      return false;
    if(!open())
      return false;
    coords.cleaR();
    atoms.clear();

    // Now we want to read data in until the record name is "END", then stop.
    int big = 0;
    int toBig = 0;
    myComment = "";
    do{
      string record(getline());
      if(equalStart("END",record))
	break;
      if(equalStart("REMARK",record) || record.empty()){
	record = removeBeginEndBlanks(record);
	if(!record.empty())
	  myComment += (myComment.empty()?"":"\n")+record;
	continue;
      }
      if(equalStart("ATOM",record)){
	record = getRightFill(record,80);
	string str = removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_RES_SEQ,PDB::PDBAtom::L_RES_SEQ));
	int resSeq = toInt(str);
	if(!isInt(str)){
	  if(str.size() == 4 && isInt(string(&str[1],&str[4])) && str[0] >= 'A' && str[0] <= 'Z'){
	    resSeq = (str[0] - 'A') * 1000 + 10000 + toInt(string(&str[1],&str[4]));
	    ++big;
	  }
	  else {
	    resSeq = -1;
	    ++toBig;
	  }
	}
	    
	atoms.push_back(PDB::PDBAtom(removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_RECORD_NAME,PDB::PDBAtom::L_RECORD_NAME)),
				     toInt(record.substr(PDB::PDBAtom::S_SERIAL,PDB::PDBAtom::L_SERIAL)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_ATOM_NAME,PDB::PDBAtom::L_ATOM_NAME)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_ALT_LOC,PDB::PDBAtom::L_ALT_LOC)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_RES_NAME,PDB::PDBAtom::L_RES_NAME)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_CHAIN_ID,PDB::PDBAtom::L_CHAIN_ID)),
				     resSeq,
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_I_CODE,PDB::PDBAtom::L_I_CODE)),
				     toReal(record.substr(PDB::PDBAtom::S_OCCUP,PDB::PDBAtom::L_OCCUP)),
				     toReal(record.substr(PDB::PDBAtom::S_TEMP,PDB::PDBAtom::L_TEMP)),
				     toInt(record.substr(PDB::PDBAtom::S_FOOT_NOTE,PDB::PDBAtom::L_FOOT_NOTE)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_SEG_ID,PDB::PDBAtom::L_SEG_ID)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_ELEMENT_SYMBOL,PDB::PDBAtom::L_ELEMENT_SYMBOL)),
				     removeBeginEndBlanks(record.substr(PDB::PDBAtom::S_CHARGE,PDB::PDBAtom::L_CHARGE)),
				     0));
	coords.push_back(Vector3D(toReal(record.substr(PDB::PDBAtom::S_X,PDB::PDBAtom::L_X)),
				  toReal(record.substr(PDB::PDBAtom::S_Y,PDB::PDBAtom::L_Y)),
				  toReal(record.substr(PDB::PDBAtom::S_Z,PDB::PDBAtom::L_Z))));
      }
      else{
	report << recoverable << "[PDB::read] Record unknow:\'"<<record<<"\'."<<endr;
      }
    }
    while (!(myFile.eof()));

    if(big > 0)
      report << hint << "[PDB::read] Found "<<big<<" X-Plor residue number(s) starting with a character."<<endr;
    if(toBig > 0){
      report << recoverable << "[PDB::read] Found "<<toBig<<" non interger/X-Plor residue number(s)."<<endr;
    }

    close();      
    return !myFile.fail();
  }

  Vector3DBlock* PDBReader::orphanCoords(){
    Vector3DBlock* tmp = myCoords;
    myCoords = NULL;
    return tmp;
  }

  std::vector<PDB::PDBAtom>* PDBReader::orphanAtoms(){
    std::vector<PDB::PDBAtom>* tmp = myAtoms;
    myAtoms = NULL;
    return tmp;
  }

  PDB PDBReader::getPDB() const{
    PDB res;
    if(myCoords != NULL)
      res.coords = (*myCoords);
    if(myAtoms != NULL)
      res.atoms = (*myAtoms);
    return res;
  }

  PDBReader& operator>>(PDBReader& pdbReader, PDB& pdb){
    pdbReader.read(pdb.coords,pdb.atoms);
    return pdbReader;
  }

  PDBReader& operator>>(PDBReader& pdbReader, Vector3DBlock& coords){
    if(pdbReader.myAtoms == NULL)
      pdbReader.myAtoms = new std::vector<PDB::PDBAtom>();
    pdbReader.read(coords,*pdbReader.myAtoms);
    return pdbReader;
  }

  PDBReader& operator>>(PDBReader& pdbReader, XYZ& xyz){
    if(pdbReader.myAtoms == NULL)
      pdbReader.myAtoms = new std::vector<PDB::PDBAtom>();
    if(pdbReader.read(xyz.coords,*pdbReader.myAtoms)){
      xyz.names.resize(xyz.coords.size());
      for(unsigned int i=0;i<xyz.coords.size();++i)
	xyz.names[i] = (*pdbReader.myAtoms)[i].elementName;
    }
    return pdbReader;
  }

}

