#include "PDBWriter.h"

#include "PDB.h"
#include "Report.h"
#include "stringutilities.h"
#include "systemutilities.h"

#include <iomanip>

using std::string;
using std::endl;
using std::setprecision;
using namespace ProtoMol::Report;

namespace ProtoMol {

  //_________________________________________________________________PDBWriter

  PDBWriter::PDBWriter():Writer(){}

  PDBWriter::PDBWriter(const std::string& filename):Writer(filename){}

  bool PDBWriter::write(const PDB& pdb){
    return write(pdb.coords,pdb.atoms);
  }

  bool PDBWriter::write(const Vector3DBlock& coords, const std::vector<PDB::PDBAtom>& atoms){
    if(!open())
      return false;

    const unsigned int count = atoms.size();
    if(coords.size() != count)
      report << error << "[PDBWriter::write]"
	     << " Size of coordinates and atoms differ."<< endr;

    myFile << "REMARKS ProtoMol (built on "<< __DATE__ << " at " << __TIME__<< ") by "<<getUserName() << endl
	   << "REMARKS This .pdb file was created by PDBWriter" << endl
	   << "REMARKS It was not manually assembled" << endl
	   << "REMARKS " <<myComment<< endl;

    int big = 0;
    int toBig = 0;
    for(unsigned int i=0;i<count;++i){
      string line(80,' ');
      const PDB::PDBAtom& a(atoms[i]);
      const Vector3D&   c(coords[i]);
      string resSeq = toString(a.residueNum);
      if(a.residueNum>=10000){
	if(a.residueNum < 36000){
	  resSeq = string(1,'A'+(a.residueNum-10000)/1000)+string(toString(a.residueNum%1000));
	  ++big;
	}
	else{
	  ++toBig;
	  resSeq = "-1";
	}
      }
      line.replace(PDB::PDBAtom::S_RECORD_NAME,
		   PDB::PDBAtom::L_RECORD_NAME,
		   getRightFill(a.elementType,PDB::PDBAtom::L_RECORD_NAME)); 
      line.replace(PDB::PDBAtom::S_SERIAL, 
		   PDB::PDBAtom::L_SERIAL, 
		   getRightFill(toString(a.elementNum),PDB::PDBAtom::L_SERIAL));
      line.replace(PDB::PDBAtom::S_ATOM_NAME, 
		   PDB::PDBAtom::L_ATOM_NAME, 
		   getRightFill(a.elementName,PDB::PDBAtom::L_ATOM_NAME));
      line.replace(PDB::PDBAtom::S_ALT_LOC, 
		   PDB::PDBAtom::L_ALT_LOC, 
		   getRightFill(a.altLoc,PDB::PDBAtom::L_ALT_LOC));
      line.replace(PDB::PDBAtom::S_RES_NAME, 
		   PDB::PDBAtom::L_RES_NAME, 
		   getRightFill(a.residueName,PDB::PDBAtom::L_RES_NAME));
      line.replace(PDB::PDBAtom::S_CHAIN_ID, 
		   PDB::PDBAtom::L_CHAIN_ID, 
		   getRightFill(a.chainID,PDB::PDBAtom::L_CHAIN_ID));
      line.replace(PDB::PDBAtom::S_RES_SEQ, 
		   PDB::PDBAtom::L_RES_SEQ, 
		   getLeftFill(resSeq,PDB::PDBAtom::L_RES_SEQ));
      line.replace(PDB::PDBAtom::S_I_CODE, 
		   PDB::PDBAtom::L_I_CODE, 
		   getRightFill(a.insertionCode,PDB::PDBAtom::L_I_CODE));
      line.replace(PDB::PDBAtom::S_X+1, 
		   PDB::PDBAtom::L_X-1, 
		   getLeftFill(toString(c.x),PDB::PDBAtom::L_X-1));
      line.replace(PDB::PDBAtom::S_Y+1, 
		   PDB::PDBAtom::L_Y-1, 
		   getLeftFill(toString(c.y),PDB::PDBAtom::L_Y-1));
      line.replace(PDB::PDBAtom::S_Z+1, 
		   PDB::PDBAtom::L_Z-1, 
		   getLeftFill(toString(c.z),PDB::PDBAtom::L_Z-1));
      line.replace(PDB::PDBAtom::S_OCCUP, 
		   PDB::PDBAtom::L_OCCUP, 
		   getLeftFill(toString(a.occupancy),PDB::PDBAtom::L_OCCUP));
      line.replace(PDB::PDBAtom::S_TEMP, 
		   PDB::PDBAtom::L_TEMP, 
		   getLeftFill(toString(a.tempFactor),PDB::PDBAtom::L_TEMP));
      line.replace(PDB::PDBAtom::S_SEG_ID, 
		   PDB::PDBAtom::L_SEG_ID, 
		   getLeftFill(a.segID,PDB::PDBAtom::L_SEG_ID));
      line.replace(PDB::PDBAtom::S_ELEMENT_SYMBOL, 
		   PDB::PDBAtom::L_ELEMENT_SYMBOL, 
		   getLeftFill(a.symbol,PDB::PDBAtom::L_ELEMENT_SYMBOL));
      line.replace(PDB::PDBAtom::S_CHARGE, 
		   PDB::PDBAtom::L_CHARGE, 
		   getLeftFill(a.charge,PDB::PDBAtom::L_CHARGE));
      myFile << line << endl;

    }
    myFile <<"END"<<endl;
    if(big > 0)
      report << hint << "[PDB::write] Wrote "<<big<<" X-Plor residue number(s) starting with a character."<<endr;
    if(toBig > 0)
      report << recoverable << "[PDB::write] Wrote "<<toBig<<" non interger/X-Plor residue number(s)."<<endr;
    close();
    return !myFile.fail();
  }

  PDBWriter& operator<<(PDBWriter& pdbWriter, const PDB& pdb){
    pdbWriter.write(pdb);
    return pdbWriter;
  }
}
